#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The constraints defining if a swift display_item is assumed a NxSpectrumSetEels concept."""

# pylint: disable=no-member,line-too-long

# releasing line-too-long restriction to avoid having line breaks in the mapping table
# made the experience that when having a widescreen working with the mapping table
# as single-line instructions is more convenient to read and parsable by human eye


NxSpectrumEels = {"IGNORE": {"fun": "load_from", "terms": "type"},
                  "IGNORE": {"fun": "load_from", "terms": "uuid"},
                  "IGNORE": {"fun": "convert_iso8601", "terms": ["created", "timezone"]},
                  "IGNORE": {"fun": "load_from", "terms": "is_sequence"},
                  "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/offset"},
                  "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/scale"},
                  "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/units"},
                  "IGNORE": {"fun": "load_from", "terms": "dimensional_calibrations"},
                  "IGNORE": {"fun": "load_from", "terms": "timezone_offset"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/header_info/header_detail"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/header_info/htype"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/header_info/series"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/auto_summation"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/beam_center_x"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/beam_center_y"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/bit_depth_image"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/bit_depth_readout"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/chi_increment"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/chi_start"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/compression"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/count_time"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/countrate_correction_applied"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/countrate_correction_count_cutoff"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/data_collection_date"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/description"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/detector_distance"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/detector_number"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/detector_readout_time"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/detector_translation"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/eiger_fw_version"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/element"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/flatfield_correction_applied"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/frame_count_time"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/frame_period"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/frame_time"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/kappa_increment"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/kappa_start"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/nimages"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/ntrigger"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/number_of_excluded_pixels"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/omega_increment"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/omega_start"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/phi_increment"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/phi_start"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/photon_energy"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/pixel_mask_applied"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/roi_mode"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/sensor_material"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/sensor_thickness"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/software_version"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/threshold_energy"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/trigger_mode"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/two_theta_increment"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/two_theta_start"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/virtual_pixel_correction_applied"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/wavelength"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/x_pixel_size"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/x_pixels_in_detector"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/y_pixel_size"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/detector_configuration/y_pixels_in_detector"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/bad_pixels"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/processing"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/flip_l_r"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/binning"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/chip_size"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/sensor_dimensions"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/readout_area"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/countrate_correction_cutoff"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/interpolate_racetracks"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/mark_saturated_pixels"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/apply_countrate_correction"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/countrate_correction_factor"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/apply_gain_correction"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/camera_processing_parameters/always_interpolate_racetracks"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/high_tension"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/defocus"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/EHT"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/MajorOL"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/StageOutA"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/StageOutB"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/StageOutX"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/StageOutY"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/StageOutZ"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/probe_ha"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/SuperFEG.^EmissionCurrent"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/LastTuneCurrent"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C10"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C12.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C12.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C21.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C21.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C23.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C23.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C30"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C32.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C32.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C34.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C34.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/ImageRonchigram/C50"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/hardware_source_id"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/hardware_source_name"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/exposure"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/binning"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/signal_type"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/hardware_source_id"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/hardware_source_name"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_id"},
                  "IGNORE": {"fun": "load_from", "terms": ["metadata/scan/center_x_nm", "metadata/scan/center_y_nm"]},
                  "IGNORE": "nm",
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/fov_nm"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/rotation"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/rotation_deg"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_context_size"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_size"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/size"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/center_nm"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/pixel_time_us"},
                  "IGNORE": "µs",
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/fov_nm"},
                  "IGNORE": "nm",
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/rotation_rad"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/external_clock_wait_time_ms"},
                  "IGNORE": "ms",
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/external_clock_mode"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/external_scan_mode"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/external_scan_ratio"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/ac_line_sync"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/ac_frame_sync"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/flyback_time_us"},
                  "IGNORE": "µs",
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/scan_id"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/scan/valid_rows"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/high_tension"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/defocus"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/EHT"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/PMTBF_gain"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/PMTDF_gain"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/StageOutA"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/StageOutB"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/StageOutX"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/StageOutY"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/StageOutZ"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C10"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C12.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C12.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C21.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C21.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C23.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C23.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C30"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C32.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C32.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C34.a"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C34.b"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C50"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C1 ConstW"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C2 ConstW"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C3 ConstW"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/PMT2_gain"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/SuperFEG.^EmissionCurrent"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/G_2Db"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/LastTuneCurrent"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/probe_ha"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/HAADF_Inner_ha"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/HAADF_Outer_ha"},
                  "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/GeometricProbeSize"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/x_shifter"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/blanker"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/x_shift_delay"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/focus"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/focus_delay"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/auto_dark_subtract"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/processing"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/blanker_delay"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/sum_frames"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/camera_hardware_source_id"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/use_multi_eels_calibration"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/shift_each_sequence_slice"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/y_shifter"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/x_units_per_ev"},
                  "UNCLEAR": "eV",
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/y_units_per_px"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/y_shift_delay"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/saturation_value"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/y_align"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.settings/stitch_spectra"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.parameters/index"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.parameters/offset_x"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.parameters/offset_y"},
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.parameters/exposure_ms"},
                  "UNCLEAR": "ms",
                  "UNCLEAR": {"fun": "load_from", "terms": "metadata/MultiAcquire.parameters/frames"},
                  "IGNORE": {"fun": "load_from", "terms": "title"},
                  "IGNORE": {"fun": "load_from", "terms": "session_id"},
                  "IGNORE": {"fun": "load_from", "terms": "session"},
                  "IGNORE": {"fun": "load_from", "terms": "category"},
                  "IGNORE": {"fun": "load_from", "terms": "version"},
                  "IGNORE": {"fun": "load_from", "terms": "modified"},
                  "IGNORE": {"fun": "load_from", "terms": "data_shape"},
                  "IGNORE": {"fun": "load_from", "terms": "data_dtype"},
                  "IGNORE": {"fun": "load_from", "terms": "collection_dimension_count"},
                  "IGNORE": {"fun": "load_from", "terms": "datum_dimension_count"},
                  "IGNORE": {"fun": "convert_iso8601", "terms": ["data_modified", "timezone"]},
                  "UNCLEAR": {"fun": "load_from", "terms": "session/site"}}

# {"fun": "convert_iso8601", "terms": ["created", "timezone"]}



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Wrapping multiple parsers for vendor files with ranging definition files."""

# pylint: disable=no-member

from typing import Dict, Any

import numpy as np

from ase.data import chemical_symbols
# ase encodes the zeroth entry as the unknown element X to have
# atom_numbers all starting with 1 up to len(chemical_symbols) - 1

from ifes_apt_tc_data_modeling.utils.utils \
    import create_isotope_vector, isotope_vector_to_nuclid_list, \
    isotope_vector_to_human_readable_name

from ifes_apt_tc_data_modeling.utils.definitions \
    import MAX_NUMBER_OF_ATOMS_PER_ION, MQ_EPSILON

from ifes_apt_tc_data_modeling.rng.rng_reader import ReadRngFileFormat

from ifes_apt_tc_data_modeling.rrng.rrng_reader import ReadRrngFileFormat

from ifes_apt_tc_data_modeling.fig.fig_reader import ReadFigTxtFileFormat

from pynxtools.dataconverter.readers.apm.utils.apm_versioning \
    import NX_APM_EXEC_NAME, NX_APM_EXEC_VERSION


def add_unknown_iontype(template: dict, entry_id: int) -> dict:
    """Add default unknown iontype."""
    # all unidentifiable ions are mapped on the unknown type
    trg = f"/ENTRY[entry{entry_id}]/atom_probe/ranging/" \
          f"peak_identification/ION[ion0]/"
    ivec = create_isotope_vector([])
    template[f"{trg}isotope_vector"] \
        = np.reshape(np.asarray(ivec, np.uint16), (1, MAX_NUMBER_OF_ATOMS_PER_ION))
    template[f"{trg}charge_state"] = np.int8(0)
    template[f"{trg}mass_to_charge_range"] \
        = np.reshape(np.asarray([0.0, MQ_EPSILON], np.float32), (1, 2))
    template[f"{trg}mass_to_charge_range/@units"] = "Da"
    nuclid_list = isotope_vector_to_nuclid_list(ivec)
    template[f"{trg}nuclid_list"] = np.asarray(nuclid_list, np.uint16)
    template[f"{trg}name"] = isotope_vector_to_human_readable_name(ivec, 0)

    return template


def add_standardize_molecular_ions(ion_lst: list, template: dict, entry_id: int) -> dict:
    """Added standard formatted molecular ion entries."""
    ion_id = 1
    trg = f"/ENTRY[entry{entry_id}]/atom_probe/ranging/peak_identification/"
    for ion in ion_lst:
        path = f"{trg}ION[ion{ion_id}]/"

        template[f"{path}isotope_vector"] = np.reshape(
            np.asarray(ion.isotope_vector.typed_value, np.uint16),
            (1, MAX_NUMBER_OF_ATOMS_PER_ION))
        template[f"{path}charge_state"] = np.int8(ion.charge_state.typed_value)
        template[f"{path}mass_to_charge_range"] \
            = np.array(ion.ranges.typed_value, np.float32)
        template[f"{path}mass_to_charge_range/@units"] = "Da"  # ion.ranges.unit
        template[f"{path}nuclid_list"] = ion.nuclid_list.typed_value
        template[f"{path}name"] = ion.name.typed_value

        path = f"{trg}ION[ion{ion_id}]/charge_state_model/"
        template[f"{path}min_abundance"] \
            = np.float64(ion.charge_state_model["min_abundance"])
        template[f"{path}min_abundance_product"] \
            = np.float64(ion.charge_state_model["min_abundance_product"])
        template[f"{path}min_half_life"] \
            = np.float64(ion.charge_state_model["min_half_life"])
        template[f"{path}min_half_life/@units"] = "s"
        template[f"{path}sacrifice_isotopic_uniqueness"] \
            = np.uint8(ion.charge_state_model["sacrifice_isotopic_uniqueness"])
        template[f"{path}isotope_matrix"] \
            = {"compress": np.array(ion.charge_state_model["isotope_matrix"],
                                    np.uint16), "strength": 1}
        template[f"{path}charge_state_vector"] \
            = {"compress": np.array(ion.charge_state_model["charge_state_vector"],
                                    np.int8), "strength": 1}
        template[f"{path}mass_vector"] \
            = {"compress": np.array(ion.charge_state_model["mass_vector"],
                                    np.float64), "strength": 1}
        template[f"{path}mass_vector/@units"] = "u"
        template[f"{path}natural_abundance_product_vector"] \
            = {"compress": np.array(ion.charge_state_model["nat_abun_prod_vector"],
                                    np.float64), "strength": 1}
        template[f"{path}min_half_life_vector"] \
            = {"compress": np.array(ion.charge_state_model["min_half_life_vector"],
                                    np.float64), "strength": 1}
        template[f"{path}min_half_life_vector/@units"] = "s"

        ion_id += 1

    trg = f"/ENTRY[entry{entry_id}]/atom_probe/ranging/"
    template[f"{trg}number_of_ion_types"] = np.uint32(ion_id)

    return template


def extract_data_from_rng_file(file_name: str, template: dict, entry_id: int) -> dict:
    """Add those required information which an RNG file has."""
    # modify the template to take into account ranging
    # ranging is currently not resolved recursively because
    # ranging(NXprocess) is a group which has a minOccurs=1, \
    #     maxOccurs="unbounded" set of possible named
    # NXion members, same case for more than one operator
    print(f"Extracting data from RNG file: {file_name}")
    rangefile = ReadRngFileFormat(file_name)

    # ion indices are on the interval [0, 256)
    assert len(rangefile.rng["molecular_ions"]) <= np.iinfo(np.uint8).max + 1, \
        "Current implementation does not support more than 256 ion types"

    add_standardize_molecular_ions(
        rangefile.rng["molecular_ions"], template, entry_id)

    return template


def extract_data_from_rrng_file(file_name: str, template: dict, entry_id) -> dict:
    """Add those required information which an RRNG file has."""
    # modify the template to take into account ranging
    # ranging is currently not resolved recursively because
    # ranging(NXprocess) is a group which has a minOccurs=1, \er
    #     maxOccurs="unbounded" set of possible named
    # NXion members, same case for more than one operator
    print(f"Extracting data from RRNG file: {file_name}")
    rangefile = ReadRrngFileFormat(file_name)

    # ion indices are on the interval [0, 256)
    assert len(rangefile.rrng["molecular_ions"]) <= np.iinfo(np.uint8).max + 1, \
        "Current implementation does not support more than 256 ion types"

    add_standardize_molecular_ions(
        rangefile.rrng["molecular_ions"], template, entry_id)

    return template


def extract_data_from_fig_txt_file(file_name: str, template: dict, entry_id) -> dict:
    """Add those required information which an transcoded Matlab figure TXT file has."""
    print(f"Extracting data from FIG.TXT file: {file_name}")
    rangefile = ReadFigTxtFileFormat(file_name)

    # ion indices are on the interval [0, 256)
    assert len(rangefile.fig["molecular_ions"]) <= np.iinfo(np.uint8).max + 1, \
        "Current implementation does not support more than 256 ion types"

    add_standardize_molecular_ions(
        rangefile.fig["molecular_ions"], template, entry_id)

    return template


class ApmRangingDefinitionsParser:  # pylint: disable=too-few-public-methods
    """Wrapper for multiple parsers for vendor specific files."""

    def __init__(self, file_name: str, entry_id: int):
        self.meta: Dict[str, Any] = {}
        self.meta["file_format"] = "none"
        self.meta["file_name"] = file_name
        self.meta["entry_id"] = entry_id
        index = file_name.lower().rfind(".")
        if index >= 0:
            mime_type = file_name.lower()[index + 1::]
            self.meta["file_format"] = mime_type

    def update_atom_types_ranging_definitions_based(self, template: dict) -> dict:
        """Update the atom_types list in the specimen based on ranging defs."""
        number_of_ion_types = 1
        prefix = f"/ENTRY[entry{self.meta['entry_id']}]/atom_probe/ranging/"
        if f"{prefix}number_of_ion_types" in template.keys():
            number_of_ion_types = template[f"{prefix}number_of_ion_types"]
        print(f"Auto-detecting elements from ranging {number_of_ion_types} ion types...")

        unique_atom_numbers = set()
        max_atom_number = len(chemical_symbols) - 1
        prefix = f"/ENTRY[entry{self.meta['entry_id']}]/atom_probe/" \
                 f"ranging/peak_identification/"
        for ion_id in np.arange(1, number_of_ion_types):
            trg = f"{prefix}ION[ion{ion_id}]/nuclid_list"
            if trg in template.keys():
                nuclid_list = template[trg][1, :]
                # second row of NXion/nuclid_list yields atom number to decode element
                for atom_number in nuclid_list:
                    if 0 < atom_number <= max_atom_number:
                        unique_atom_numbers.add(atom_number)
        print(f"Unique atom numbers are: {list(unique_atom_numbers)}")
        unique_elements = set()
        for atom_number in unique_atom_numbers:
            unique_elements.add(chemical_symbols[atom_number])
        print(f"Unique elements are: {list(unique_elements)}")

        atom_types_str = ", ".join(list(unique_elements))
        if atom_types_str != "":
            trg = f"/ENTRY[entry{self.meta['entry_id']}]/specimen/"
            template[f"{trg}atom_types"] = atom_types_str

        return template

    def report(self, template: dict) -> dict:
        """Copy data from self into template the appdef instance.

        Paths in template are prefixed by prefix and have to be compliant
        with the application definition.
        """
        # resolve the next two program references more informatively
        trg = f"/ENTRY[entry{self.meta['entry_id']}]/atom_probe/ranging/"
        template[f"{trg}maximum_number_of_atoms_per_molecular_ion"] = np.uint32(32)

        # mass_to_charge_distribution will be filled by default plot
        # background_quantification data are not available in RNG/RRNG files
        # peak_search_and_deconvolution data are not available in RNG/RRNG files

        trg = f"/ENTRY[entry{self.meta['entry_id']}]/atom_probe/" \
              f"ranging/peak_identification/"
        template[f"{trg}PROGRAM[program1]/program"] = NX_APM_EXEC_NAME
        template[f"{trg}PROGRAM[program1]/program/@version"] = NX_APM_EXEC_VERSION

        add_unknown_iontype(template, self.meta["entry_id"])

        if self.meta["file_name"] != "" and self.meta["file_format"] != "none":
            if self.meta["file_format"] == "rng":
                extract_data_from_rng_file(
                    self.meta["file_name"],
                    template,
                    self.meta["entry_id"])
            elif self.meta["file_format"] == "rrng":
                extract_data_from_rrng_file(
                    self.meta["file_name"],
                    template,
                    self.meta["entry_id"])
            elif self.meta["file_format"] == "txt":
                extract_data_from_fig_txt_file(
                    self.meta["file_name"],
                    template,
                    self.meta["entry_id"])
            else:
                trg = f"/ENTRY[entry{self.meta['entry_id']}]/atom_probe/ranging/"
                template[f"{trg}number_of_ion_types"] = 1
        else:
            trg = f"/ENTRY[entry{self.meta['entry_id']}]/atom_probe/ranging/"
            template[f"{trg}number_of_ion_types"] = 1

        self.update_atom_types_ranging_definitions_based(template)

        return template



#!/usr/bin/env python
"""Writes a NeXus HDF5 file using h5py and numpy"""

from pathlib import Path
from re import X
import numpy

from nexusformat.nexus import NXdata, NXentry, NXfield, nxopen

print("Write a NeXus HDF5 file")
fileName = "simple_example_basic.nexus.hdf5"

# load data from two column format
data_filename = str(Path(__file__).absolute().parent.parent / "simple_example.dat")
data = numpy.loadtxt(data_filename).T
mr_arr = data[0]
i00_arr = numpy.asarray(data[1], "int32")

# create the HDF5 NeXus file
with nxopen(fileName, "w") as f:

    # create the NXentry group
    f["entry"] = NXentry()
    f["entry/title"] = "1-D scan of I00 v. mr"

    # create the NXdata group
    x = NXfield(mr_arr, name="mr", units="degrees", long_name="USAXS mr (degrees)")
    y = NXfield(i00_arr, name="I00", units="counts",
                long_name="USAXS I00 (counts)")
    f["entry/mr_scan"] = NXdata(y, x)

print("wrote file:", fileName)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""General metadata object connecting units and values for a quantity."""

# pylint: disable=no-member

from typing import Dict

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_versioning \
    import NX_EM_ADEF_NAME, NX_EM_ADEF_VERSION, \
    NX_EM_EXEC_NAME, NX_EM_EXEC_VERSION


class NxObject:  # pylint: disable=too-few-public-methods
    """An object in a graph e.g. a field or group in NeXus."""

    def __init__(self,
                 name: str = None,
                 unit: str = None,
                 dtype=str,
                 value=None,
                 **kwargs):
        if name is not None:
            assert name != "", "Argument name needs to be a non-empty string !"
        if unit is not None:
            assert unit != "", "Argument unit needs to be a non-empty string !"
        assert dtype is not None, "Argument dtype must not be None !"
        if dtype is not None:
            assert isinstance(dtype, type), \
                "Argument dtype needs a valid, ideally numpy, datatype !"
        # ##MK::if value is not None:
        self.is_a = "NXobject"
        self.is_attr = False  # if True indicates object is attribute
        self.doc = ""  # docstring
        self.name = name  # name of the field
        self.unit = unit  # not unit category but actual unit
        # use special values "unitless" for NX_UNITLESS (e.g. 1) and
        # "dimensionless" for NX_DIMENSIONLESS (e.g. 1m / 1m)
        self.dtype = dtype  # use np.dtype if possible
        if value is None or dtype is str:
            self.unit = "unitless"
        if value is not None:
            self.value = value
        else:
            self.value = None
        # value should be a numpy scalar, tensor, or string if possible
        if "is_attr" in kwargs:
            assert isinstance(kwargs["is_attr"], bool), \
                "Kwarg is_attr needs to be a boolean !"
            self.is_attr = kwargs["is_attr"]

    def __repr__(self):
        """Report values."""
        return f'''Name: {self.name}, unit: {self.unit}, dtype: {self.dtype}'''


class NxEmUser:  # pylint: disable=too-few-public-methods
    """An object representing a user, typically a human."""

    def __init__(self):
        self.meta: Dict[str, NxObject] = {}
        self.meta["name"] = NxObject()
        self.meta["affiliation"] = NxObject()
        self.meta["address"] = NxObject()
        self.meta["email"] = NxObject()
        self.meta["orcid"] = NxObject()
        self.meta["telephone_number"] = NxObject()
        self.meta["role"] = NxObject()
        self.meta["social_media_name"] = NxObject()
        self.meta["social_media_platform"] = NxObject()

    def report(self, prefix: str, template: dict) -> dict:
        """Copy data from self into template the appdef instance.

        Paths in template are prefixed by prefix and have to be compliant
        with the application definition.
        """
        template[f"{prefix}/address"] = self.meta["address"].value
        template[f"{prefix}/affiliation"] = self.meta["affiliation"].value
        template[f"{prefix}/email"] = self.meta["email"].value
        template[f"{prefix}/name"] = self.meta["name"].value
        template[f"{prefix}/orcid"] = self.meta["orcid"].value
        template[f"{prefix}/role"] = self.meta["role"].value
        template[f"{prefix}/social_media_name"] \
            = self.meta["social_media_name"].value
        template[f"{prefix}/social_media_platform"] \
            = self.meta["social_media_platform"].value
        template[f"{prefix}/telephone_number"] \
            = self.meta["telephone_number"].value
        return template


class NxEmSample:  # pylint: disable=too-few-public-methods
    """An object representing a sample."""

    def __init__(self):
        self.meta: Dict[str, NxObject] = {}
        self.meta["method"] = NxObject(value="experimental")
        self.meta["name"] = NxObject()
        self.meta["sample_history"] = NxObject()
        self.meta["preparation_date"] = NxObject()
        self.meta["short_title"] = NxObject()
        self.meta["atom_types"] = NxObject(value=[])
        self.meta["thickness"] = NxObject()
        self.meta["description"] = NxObject()

    def report(self, prefix: str, template: dict) -> dict:
        """Copy data from self into template the appdef instance.

        Paths in template are prefixed by prefix and have to be compliant
        with the application definition.
        """
        template[f"{prefix}/method"] = self.meta["method"].value
        template[f"{prefix}/name"] = self.meta["name"].value
        template[f"{prefix}/sample_history"] = self.meta["sample_history"].value
        template[f"{prefix}/preparation_date"] = self.meta["preparation_date"].value
        template[f"{prefix}/short_title"] = self.meta["short_title"].value
        template[f"{prefix}/atom_types"] = self.meta["atom_types"].value
        template[f"{prefix}/thickness"] = self.meta["thickness"].value
        template[f"{prefix}/thickness/@units"] = self.meta["thickness"].unit
        template[f"{prefix}/description"] = self.meta["description"].value
        return template


class NxEmAppDefHeader:  # pylint: disable=too-few-public-methods
    """An object representing the typical header of nexus-fairmat appdefs."""

    def __init__(self):
        self.meta: Dict[str, NxObject] = {}
        self.meta["version"] \
            = NxObject(value=NX_EM_ADEF_VERSION, is_attr=True)
        self.meta["definition"] \
            = NxObject(value=NX_EM_ADEF_NAME)
        self.meta["experiment_identifier"] = NxObject()
        self.meta["experiment_description"] = NxObject()
        self.meta["start_time"] = NxObject()
        self.meta["end_time"] = NxObject()
        self.meta["program"] = NxObject(value=NX_EM_EXEC_NAME)
        self.meta["program_version"] \
            = NxObject(value=NX_EM_EXEC_VERSION, is_attr=True)
        self.meta["experiment_documentation"] = NxObject()
        self.meta["thumbnail"] = NxObject()
        self.meta["thumbnail_type"] = NxObject()

    def report(self, prefix: str, template: dict) -> dict:
        """Copy data from self into template the appdef instance.

        Paths in template are prefixed by prefix and have to be compliant
        with the application definition.
        """
        template[f"{prefix}/@version"] = self.meta["version"].value
        template[f"{prefix}/definition"] = self.meta["definition"].value
        template[f"{prefix}/experiment_identifier"] \
            = self.meta["experiment_identifier"].value
        template[f"{prefix}/experiment_description"] \
            = self.meta["experiment_description"].value
        template[f"{prefix}/start_time"] = self.meta["start_time"].value
        template[f"{prefix}/end_time"] = self.meta["end_time"].value
        template[f"{prefix}/program"] = self.meta["program"].value
        template[f"{prefix}/program/@version"] \
            = self.meta["program_version"].value
        template[f"{prefix}/experiment_documentation"] \
            = self.meta["experiment_documentation"].value
        template[f"{prefix}/thumbnail"] \
            = self.meta["thumbnail"].value
        template[f"{prefix}/thumbnail/@type"] \
            = self.meta["thumbnail_type"].value
        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This modules creates a formula parser"""
import os
from operator import add, sub, mul, truediv, neg
from typing import Dict
import numpy as np
from scipy.special import dawsn  # pylint: disable=no-name-in-module
import scipy.constants as sc
from lark import Lark, Transformer, v_args


@v_args(inline=True)
class FormulaTransformer(Transformer):
    """Transformer class for parsing formulas"""

    single_params: Dict[str, float]
    repeated_params: Dict[str, np.ndarray]
    no_repeated_params: int

    number = float
    add = add
    sub = sub
    mul = mul
    div = truediv
    neg = neg
    power = pow

    def _check_and_set(self, repeated_params):
        if not repeated_params:
            self.no_repeated_params = 0
            return

        params = iter(repeated_params.items())
        length = len(next(params)[1])
        for name, param in params:
            if not isinstance(param, np.ndarray):
                raise TypeError(
                    f"Expected {name} to be of type numpy ndarray but found type {type(param)}"
                )
            if length != len(param):
                raise ValueError("Repeated parameters must have all the same length.")

        self.no_repeated_params = length

    def _check_and_set_single(self, single_params):
        for name, param in single_params.items():
            if not isinstance(param, float):
                raise TypeError(
                    f"Expected {name} to be of type float but found type {type(param)}."
                )
        self.single_params = single_params

    def __init__(
        self,
        x_axis_name: str,
        x_axis_values: np.ndarray,
        single_params: Dict[str, float],
        repeated_params: Dict[str, np.ndarray],
    ):
        super().__init__()
        if not isinstance(x_axis_name, str):
            raise TypeError("x_axis_name must be a string.")

        if not isinstance(x_axis_values, np.ndarray):
            raise TypeError("x_axis_values must be a numpy array.")

        self.x_axis_name = x_axis_name
        self.x_axis_values = x_axis_values

        self._check_and_set(repeated_params)
        self._check_and_set_single(single_params)

        self.repeated_params = repeated_params
        self.single_params = single_params

    def eps(self, inp):
        """Return an epsilon type formula"""
        return "eps", inp

    # pylint: disable=invalid-name
    def n(self, inp):
        """Return an index type formula"""
        return "n", inp

    def kkr_term(self, term):
        """Calculate the kramers kronig transformation on the function"""
        raise NotImplementedError("kkr transformation not yet implemented")

    def func(self, name, val):
        """Evaluates a function"""
        names = {
            "sin": np.sin,
            "cos": np.cos,
            "tan": np.tan,
            "sqrt": np.emath.sqrt,
            "dawsn": dawsn,
            "ln": np.log,
            "log": np.log10,
            "heaviside": np.heaviside,
        }

        if name in names:
            return names[name](val)

        raise ValueError(f"Unknown function: {name}")

    def builtin(self, name):
        """Returns the values for builtin tokens"""
        names = {
            "1j": 1j,
            "pi": sc.pi,
            "eps_0": sc.epsilon_0,
            "hbar": sc.hbar,
            "h": sc.h,
            "c": sc.c,
        }

        if name in names:
            return names[name]

        raise ValueError(f"Unknown constant: {name}")

    def sum_expr(self, expr):
        """Sum an expression"""
        return expr.sum(axis=1)

    def single_param_name(self, name):
        """Return a parameter inside a non-repeated section"""
        if name == self.x_axis_name:
            return self.x_axis_values

        if name in self.single_params:
            return self.single_params[name]

        raise ValueError(f"No such parameter {name}")

    def param_name(self, name):
        """Return a parameter inside a repeated section"""
        if name == self.x_axis_name:
            return np.einsum(
                "i,j->ij", self.x_axis_values, np.ones(self.no_repeated_params)
            )

        if name in self.single_params:
            return self.single_params[name]

        if name in self.repeated_params:
            return self.repeated_params[name]

        raise ValueError(f"No such parameter {name}")


def formula_parser(transformer=None):
    """Creates a formula parser.

    Args:
        transformer (lark.Transformer, optional):
            A transformer to use with this parsing.
            If set to None it generates the AST.
            Defaults to None.

    Returns:
        lark.Lark: The parsing object.
    """
    __location__ = os.path.realpath(
        os.path.join(os.getcwd(), os.path.dirname(__file__))
    )

    with open(
        f"{__location__}/dispersion_function_grammar.lark", encoding="utf-8"
    ) as formula_grammar:
        if transformer is None:
            return Lark(
                formula_grammar,
                start="assignment",
                parser="lalr",
            )

        return Lark(
            formula_grammar, start="assignment", parser="lalr", transformer=transformer
        )


def transformation_formula_parser(
    x_axis_name: str,
    x_axis_values: np.ndarray,
    single_params: Dict[str, float],
    repeated_params: Dict[str, np.ndarray],
):
    """Creates a formula parser with the given parameters.
    This parser is automatically evaluated based on the given parameters.
    This is done directly whitout creating the abstract syntax tree.

    Args:
        x_axis_name (str): The name of the x-axis to use replace in the formula string.
        x_axis_values (np.ndarray): The value array of the x-axis to replace in the formula string.
        single_params (Dict[str, float]):
            A dictionary of parameter name and value of single parameters.
        repeated_params (Dict[str, np.ndarray]):
            A dictionary of parameter name and values of repeated paremters.

    Returns:
        str, np.ndarray:
            The evaluated formula.
            The first tuple element indicates the type of dispersion
            and may either be `n` for refractive index or `eps` for dielectric function.
            The second element is the evluated formula one point for each x-axis input point.
    """
    return formula_parser(
        FormulaTransformer(
            x_axis_name=x_axis_name,
            x_axis_values=x_axis_values,
            single_params=single_params,
            repeated_params=repeated_params,
        )
    )



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Wrapping multiple parsers for vendor files with reconstructed dataset files."""

# pylint: disable=no-member

import numpy as np

from ifes_apt_tc_data_modeling.apt.apt6_reader import ReadAptFileFormat

from ifes_apt_tc_data_modeling.pos.pos_reader import ReadPosFileFormat

from ifes_apt_tc_data_modeling.epos.epos_reader import ReadEposFileFormat


def extract_data_from_pos_file(file_name: str, prefix: str, template: dict) -> dict:
    """Add those required information which a POS file has."""
    print(f"Extracting data from POS file: {file_name}")
    posfile = ReadPosFileFormat(file_name)

    trg = f"{prefix}reconstruction/"
    xyz = posfile.get_reconstructed_positions()
    template[f"{trg}reconstructed_positions"] \
        = {"compress": np.array(xyz.typed_value, np.float32), "strength": 1}
    template[f"{trg}reconstructed_positions/@units"] = xyz.unit
    del xyz

    trg = f"{prefix}mass_to_charge_conversion/"
    m_z = posfile.get_mass_to_charge_state_ratio()
    template[f"{trg}mass_to_charge"] \
        = {"compress": np.array(m_z.typed_value, np.float32), "strength": 1}
    template[f"{trg}mass_to_charge/@units"] = m_z.unit
    del m_z
    return template


def extract_data_from_epos_file(file_name: str, prefix: str, template: dict) -> dict:
    """Add those required information which an ePOS file has."""
    print(f"Extracting data from EPOS file: {file_name}")
    eposfile = ReadEposFileFormat(file_name)

    trg = f"{prefix}reconstruction/"
    xyz = eposfile.get_reconstructed_positions()
    template[f"{trg}reconstructed_positions"] \
        = {"compress": np.array(xyz.typed_value, np.float32), "strength": 1}
    template[f"{trg}reconstructed_positions/@units"] = xyz.unit
    del xyz

    trg = f"{prefix}mass_to_charge_conversion/"
    m_z = eposfile.get_mass_to_charge_state_ratio()
    template[f"{trg}mass_to_charge"] \
        = {"compress": np.array(m_z.typed_value, np.float32), "strength": 1}
    template[f"{trg}mass_to_charge/@units"] = m_z.unit
    del m_z

    # there are inconsistencies in the literature as to which units these
    # quantities have, so we skip exporting the following quantities for now
    # the following source code has not been tested with the current NXapm version
    # but should be kept for making additions in the future easier
    # -->
    # trg = f"{prefix}voltage_and_bowl_correction/"
    # raw_tof = eposfile.get_raw_time_of_flight()
    # template[f"{trg}raw_tof"] = raw_tof.value
    # template[f"{trg}raw_tof/@units"] = raw_tof.unit
    # # this somehow calibrated ToF is not available from an EPOS file
    # template[f"{trg}calibrated_tof"] = raw_tof.value
    # template[f"{trg}calibrated_tof/@units"] = raw_tof.unit
    # # is this really a raw ToF, if so, raw wrt to what?
    # # needs clarification from Cameca/AMETEK how this is internally computed
    # # especially when scientists write APT files and transcode them
    # # to EPOS using APSuite
    # del raw_tof

    # trg = f"{prefix}pulser/"
    # dc_voltage = eposfile.get_standing_voltage()
    # template[f"{trg}standing_voltage"] = dc_voltage.value
    # template[f"{trg}standing_voltage/@units"] = dc_voltage.unit
    # del dc_voltage

    # pu_voltage = eposfile.get_pulse_voltage()
    # template[f"{trg}pulsed_voltage"] = pu_voltage.value
    # template[f"{trg}pulsed_voltage/@units"] = pu_voltage.unit
    # del pu_voltage

    # trg = f"{prefix}ion_impact_positions/"
    # hit_positions = eposfile.get_hit_positions()
    # template[f"{trg}hit_positions"] = hit_positions.value
    # template[f"{trg}hit_positions/@units"] = hit_positions.unit
    # del hit_positions

    # trg = f"{prefix}hit_multiplicity/"
    # # little bit more discussion with e.g. F. M. M. at MPIE required

    # # currently npulses is "number of pulses since last event detected"
    # npulses = eposfile.get_number_of_pulses()
    # template[f"{trg}hit_multiplicity"] = npulses.value
    # template[f"{trg}hit_multiplicity/@units"] = npulses.unit
    # del npulses

    # ions_per_pulse = eposfile.get_ions_per_pulse()
    # # currently ions_per_pulse is "ions per pulse, 0 after the first ion"
    # template[f"{trg}pulses_since_last_ion"] = ions_per_pulse.value
    # template[f"{trg}pulses_since_last_ion/@units"] \
    # = ions_per_pulse.unit
    # del ions_per_pulse
    # -->

    return template


def extract_data_from_apt_file(file_name: str, prefix: str, template: dict) -> dict:
    """Add those required information which a APT file has."""
    print(f"Extracting data from APT file: {file_name}")
    aptfile = ReadAptFileFormat(file_name)

    trg = f"{prefix}reconstruction/"
    xyz = aptfile.get_named_quantity("Position")
    template[f"{trg}reconstructed_positions"] \
        = {"compress": np.array(xyz.typed_value, np.float32), "strength": 1}
    template[f"{trg}reconstructed_positions/@units"] = xyz.unit
    del xyz

    trg = f"{prefix}mass_to_charge_conversion/"
    m_z = aptfile.get_named_quantity("Mass")
    template[f"{trg}mass_to_charge"] \
        = {"compress": np.array(m_z.typed_value, np.float32), "strength": 1}
    template[f"{trg}mass_to_charge/@units"] = m_z.unit
    del m_z

    # all less explored optional branches in an APT6 file can also already
    # be accessed via the aptfile.get_named_quantity function
    # but it needs to be checked if this returns reasonable values
    # and specifically what these values logically mean, interaction with
    # Cameca as well as the community is vital here
    return template


class ApmReconstructionParser:  # pylint: disable=too-few-public-methods
    """Wrapper for multiple parsers for vendor specific files."""

    def __init__(self, file_name: str, entry_id: int):
        self.file_format = "none"
        self.file_name = file_name
        index = file_name.lower().rfind(".")
        if index >= 0:
            mime_type = file_name.lower()[index + 1::]
            if mime_type in ["pos", "epos", "apt"]:
                self.file_format = mime_type
        self.entry_id = entry_id

    def report(self, template: dict) -> dict:
        """Copy data from self into template the appdef instance.

        Paths in template are prefixed by prefix and have to be compliant
        with the application definition.
        """
        prfx = f"/ENTRY[entry{self.entry_id}]/atom_probe/"
        if self.file_name != "" and self.file_format != "none":
            if self.file_format == "pos":
                extract_data_from_pos_file(
                    self.file_name, prfx, template)
            if self.file_format == "epos":
                extract_data_from_epos_file(
                    self.file_name, prfx, template)
            if self.file_format == "apt":
                extract_data_from_apt_file(
                    self.file_name, prfx, template)
        return template



#!/usr/bin/env python3
"""Functionalities shared across different parsers."""

# -*- coding: utf-8 -*-
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=E1101, R0801

# import git


def get_repo_last_commit() -> str:
    """Identify the last commit to the repository."""
    # repo = git.Repo(search_parent_directories=True)
    # sha = str(repo.head.object.hexsha)
    # if sha != "":
    #    return sha
    # currently update-north-markus branch on nomad-FAIR does not pick up
    # git even though git in the base image and gitpython in pynxtools deps
    return "unknown git commit id or unable to parse git reverse head"


class NxObject:  # pylint: disable=R0903
    """An object in a graph e.g. a field or group in NeXus."""

    def __init__(self,
                 name: str = None,
                 unit: str = None,
                 dtype=str,
                 value=None,
                 **kwargs):
        if name is not None:
            assert name != "", "Argument name needs to be a non-empty string !"
        if unit is not None:
            assert unit != "", "Argument unit needs to be a non-empty string !"
        assert dtype is not None, "Argument dtype must not be None !"
        if dtype is not None:
            assert isinstance(dtype, type), \
                "Argument dtype needs a valid, ideally numpy, datatype !"
        # ##MK::if value is not None:
        self.is_a = "NXobject"
        self.is_attr = False  # if True indicates object is attribute
        self.doc = ""  # docstring
        self.name = name  # name of the field
        self.unit = unit  # not unit category but actual unit
        # use special values "unitless" for NX_UNITLESS (e.g. 1) and
        # "dimensionless" for NX_DIMENSIONLESS (e.g. 1m / 1m)
        self.dtype = dtype  # use np.dtype if possible
        if value is None or dtype is str:
            self.unit = "unitless"
        if value is not None:
            self.value = value
        else:
            self.value = None
        # value should be a numpy scalar, tensor, or string if possible
        if "is_attr" in kwargs:
            assert isinstance(kwargs["is_attr"], bool), \
                "Kwarg is_attr needs to be a boolean !"
            self.is_attr = kwargs["is_attr"]

    def print(self):
        """Report values."""
        print("name: ")
        print(str(self.name))
        print("unit:")
        print(str(self.unit))
        print("dtype: ")
        print(self.dtype)

# test = NxObject(name="test", unit="baud", dtype=np.uint32, value=32000)
# test.print()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This script runs the conversion routine using a selected reader and write out a NeXus file."""
import glob
import importlib.machinery
import importlib.util
import logging
import os
import sys
from typing import List, Tuple
import xml.etree.ElementTree as ET

import click
import yaml

from pynxtools.dataconverter.readers.base.reader import BaseReader
from pynxtools.dataconverter import helpers
from pynxtools.dataconverter.writer import Writer
from pynxtools.dataconverter.template import Template
from pynxtools.nexus import nexus

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points


logger = logging.getLogger(__name__)  # pylint: disable=C0103
UNDOCUMENTED = 9
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def get_reader(reader_name) -> BaseReader:
    """Helper function to get the reader object from it's given name"""
    path_prefix = f"{os.path.dirname(__file__)}{os.sep}" if os.path.dirname(__file__) else ""
    path = os.path.join(path_prefix, "readers", reader_name, "reader.py")
    spec = importlib.util.spec_from_file_location("reader.py", path)
    try:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore[attr-defined]
    except FileNotFoundError as exc:
        # pylint: disable=unexpected-keyword-arg
        importlib_module = entry_points(group='pynxtools.reader')
        if (
            importlib_module
            and reader_name in map(lambda ep: ep.name, entry_points(group='pynxtools.reader'))
        ):
            return importlib_module[reader_name].load()
        raise ValueError(f"The reader, {reader_name}, was not found.") from exc
    return module.READER  # type: ignore[attr-defined]


def get_names_of_all_readers() -> List[str]:
    """Helper function to populate a list of all available readers"""
    path_prefix = f"{os.path.dirname(__file__)}{os.sep}" if os.path.dirname(__file__) else ""
    files = sorted(glob.glob(os.path.join(path_prefix, "readers", "*", "reader.py")))
    all_readers = []
    for file in files:
        if f"{os.sep}base{os.sep}" not in file:
            index_of_readers_folder_name = file.rindex(f"readers{os.sep}") + len(f"readers{os.sep}")
            index_of_last_path_sep = file.rindex(os.sep)
            all_readers.append(file[index_of_readers_folder_name:index_of_last_path_sep])
    plugins = list(map(lambda ep: ep.name, entry_points(group='pynxtools.reader')))
    return all_readers + plugins


# pylint: disable=too-many-arguments,too-many-locals
def convert(input_file: Tuple[str, ...],
            reader: str,
            nxdl: str,
            output: str,
            generate_template: bool = False,
            fair: bool = False,
            undocumented: bool = False,
            **kwargs):
    """The conversion routine that takes the input parameters and calls the necessary functions."""
    # Reading in the NXDL and generating a template
    definitions_path = nexus.get_nexus_definitions_path()
    if nxdl == "NXtest":
        nxdl_path = os.path.join(
            f"{os.path.abspath(os.path.dirname(__file__))}/../../",
            "tests", "data", "dataconverter", "NXtest.nxdl.xml")
    elif nxdl == "NXroot":
        nxdl_path = os.path.join(definitions_path, "base_classes", "NXroot.nxdl.xml")
    else:
        nxdl_path = os.path.join(definitions_path, "contributed_definitions", f"{nxdl}.nxdl.xml")
        if not os.path.exists(nxdl_path):
            nxdl_path = os.path.join(definitions_path, "applications", f"{nxdl}.nxdl.xml")
        if not os.path.exists(nxdl_path):
            raise FileNotFoundError(f"The nxdl file, {nxdl}, was not found.")

    nxdl_root = ET.parse(nxdl_path).getroot()

    if undocumented:
        logger.setLevel(UNDOCUMENTED)

    template = Template()
    helpers.generate_template_from_nxdl(nxdl_root, template)
    if generate_template:
        logger.info(template)
        return

    # Setting up all the input data
    if isinstance(input_file, str):
        input_file = (input_file,)
    bulletpoint = "\n\u2022 "
    logger.info("Using %s reader to convert the given files: %s ",
                reader,
                bulletpoint.join((" ", *input_file)))

    data_reader = get_reader(reader)
    if not (nxdl in data_reader.supported_nxdls or "*" in data_reader.supported_nxdls):
        raise NotImplementedError("The chosen NXDL isn't supported by the selected reader.")

    data = data_reader().read(  # type: ignore[operator]
        template=Template(template),
        file_paths=input_file,
        **kwargs,
    )
    helpers.validate_data_dict(template, data, nxdl_root)
    if fair and data.undocumented.keys():
        logger.warning("There are undocumented paths in the template. This is not acceptable!")
        return

    for path in data.undocumented.keys():
        if "/@default" in path:
            continue
        logger.log(
            UNDOCUMENTED,
            "The path, %s, is being written but has no documentation.",
            path
        )
    helpers.add_default_root_attributes(data=data, filename=os.path.basename(output))
    Writer(data=data, nxdl_path=nxdl_path, output_path=output).write()

    logger.info("The output file generated: %s", output)


def parse_params_file(params_file):
    """Parses the parameters from a given dictionary and returns them"""
    params = yaml.load(params_file, Loader=yaml.Loader)['dataconverter']
    for param in list(params.keys()):
        params[param.replace("-", "_")] = params.pop(param)
    return params


@click.command()
@click.option(
    '--input-file',
    default=[],
    multiple=True,
    help='The path to the input data file to read. (Repeat for more than one file.)'
)
@click.option(
    '--reader',
    default='json_map',
    type=click.Choice(get_names_of_all_readers(), case_sensitive=False),
    help='The reader to use. default="example"'
)
@click.option(
    '--nxdl',
    default=None,
    required=False,
    help='The name of the NXDL file to use without extension.'
)
@click.option(
    '--output',
    default='output.nxs',
    help='The path to the output NeXus file to be generated.'
)
@click.option(
    '--generate-template',
    is_flag=True,
    default=False,
    help='Just print out the template generated from given NXDL file.'
)
@click.option(
    '--fair',
    is_flag=True,
    default=False,
    help='Let the converter know to be stricter in checking the documentation.'
)
@click.option(
    '--params-file',
    type=click.File('r'),
    default=None,
    help='Allows to pass a .yaml file with all the parameters the converter supports.'
)
@click.option(
    '--undocumented',
    is_flag=True,
    default=False,
    help='Shows a log output for all undocumented fields'
)
@click.option(
    '--mapping',
    help='Takes a <name>.mapping.json file and converts data from given input files.'
)
# pylint: disable=too-many-arguments
def convert_cli(input_file: Tuple[str, ...],
                reader: str,
                nxdl: str,
                output: str,
                generate_template: bool,
                fair: bool,
                params_file: str,
                undocumented: bool,
                mapping: str):
    """The CLI entrypoint for the convert function"""
    if params_file:
        try:
            convert(**parse_params_file(params_file))
        except TypeError as exc:
            sys.tracebacklimit = 0
            raise TypeError(("Please make sure you have the following entries in your "
                             "parameter file:\n\n# NeXusParser Parameter File - v0.0.1"
                             "\n\ndataconverter:\n\treader: value\n\tnxdl: value\n\tin"
                             "put-file: value")) from exc
    else:
        if nxdl is None:
            sys.tracebacklimit = 0
            raise IOError("\nError: Please supply an NXDL file with the option:"
                          " --nxdl <path to NXDL>")
        if mapping:
            reader = "json_map"
            if mapping:
                input_file = input_file + tuple([mapping])
        convert(input_file, reader, nxdl, output, generate_template, fair, undocumented)


if __name__ == '__main__':
    convert_cli()  # pylint: disable=no-value-for-parameter



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Perkin Ellmer transmission file reader implementation for the DataConverter."""
from inspect import isfunction
from typing import Callable, List, Any, Dict
import pandas as pd

from pynxtools.dataconverter.readers.json_yml.reader import YamlJsonReader
import pynxtools.dataconverter.readers.transmission.metadata_parsers as mpars
from pynxtools.dataconverter.readers.utils import parse_json, parse_yml, handle_objects


# Dictionary mapping metadata in the asc file to the paths in the NeXus file.
# The entry can either be a function with one list parameter
# which is executed to fill the specific path or an
# integer which is used to get the value at the index of the metadata.
# If the value is a str this string gets inputed at the path.
METADATA_MAP: Dict[str, Any] = {
    "/ENTRY[entry]/SAMPLE[sample]/name": 8,
    "/ENTRY[entry]/start_time": mpars.read_start_date,
    "/ENTRY[entry]/instrument/sample_attenuator/attenuator_transmission":
        mpars.read_sample_attenuator,
    "/ENTRY[entry]/instrument/ref_attenuator/attenuator_transmission":
        mpars.read_ref_attenuator,
    "/ENTRY[entry]/instrument/common_beam_mask/y_gap": 45,
    "/ENTRY[entry]/instrument/polarizer": 48,
    "/ENTRY[entry]/instrument/common_beam_depolarizer":
        mpars.is_depolarizer_on,
    "/ENTRY[entry]/instrument/spectrometer/GRATING[grating]/wavelength_range":
        mpars.read_uv_monochromator_range,
    "/ENTRY[entry]/instrument/spectrometer/GRATING[grating1]/wavelength_range":
        mpars.read_visir_monochromator_range,
    "/ENTRY[entry]/instrument/SOURCE[source]/type": "D2",
    "/ENTRY[entry]/instrument/SOURCE[source]/wavelength_range": mpars.get_d2_range,
    "/ENTRY[entry]/instrument/SOURCE[source1]/type": "halogen",
    "/ENTRY[entry]/instrument/SOURCE[source1]/wavelength_range": mpars.get_halogen_range,
}
# Dictionary to map value during the yaml eln reading
# This is typically a mapping from ELN signifier to NeXus path
CONVERT_DICT: Dict[str, str] = {"Sample": "SAMPLE[sample]", "unit": "@units"}
# Dictionary to map nested values during the yaml eln reading
# This is typically a mapping from nested ELN signifiers to NeXus group
REPLACE_NESTED: Dict[str, str] = {
    "Sample[sample]/experiment_identifier": "experiment_identifier"
}


def data_to_template(data: pd.DataFrame) -> Dict[str, Any]:
    """Builds the data entry dict from the data in a pandas dataframe

    Args:
        data (pd.DataFrame): The dataframe containing the data.

    Returns:
        Dict[str, Any]: The dict with the data paths inside NeXus.
    """
    template: Dict[str, Any] = {}
    template["/ENTRY[entry]/data/@axes"] = "wavelength"
    template["/ENTRY[entry]/data/type"] = "transmission"
    template["/ENTRY[entry]/data/@signal"] = "transmission"
    template["/ENTRY[entry]/data/wavelength"] = data.index.values
    template["/ENTRY[entry]/instrument/spectrometer/wavelength"] = data.index.values
    template["/ENTRY[entry]/data/wavelength/@units"] = "nm"
    template["/ENTRY[entry]/data/transmission"] = data.values[:, 0]
    template["/ENTRY[entry]/instrument/measured_data"] = data.values

    return template


def parse_detector_line(line: str, convert: Callable[[str], Any] = None) -> List[Any]:
    """Parses a detector line from the asc file.

    Args:
        line (str): The line to parse.

    Returns:
        List[Any]: The list of detector settings.
    """
    if convert is None:

        def convert(val):
            return val

    return [convert(s.split("/")[-1]) for s in line.split()]


# pylint: disable=too-many-arguments
def convert_detector_to_template(
    det_type: str,
    slit: str,
    time: float,
    gain: float,
    det_idx: int,
    wavelength_range: List[float],
) -> Dict[str, Any]:
    """Writes the detector settings to the template.

    Args:
        det_type (str): The detector type.
        slit (float): The slit width.
        time (float): The exposure time.
        gain (str): The gain setting.

    Returns:
        Dict[str, Any]: The dictionary containing the data readout from the asc file.
    """
    if det_idx == 0:
        path = "/ENTRY[entry]/instrument/DETECTOR[detector]"
    else:
        path = f"/ENTRY[entry]/instrument/DETECTOR[detector{det_idx}]"
    template: Dict[str, Any] = {}
    template[f"{path}/type"] = det_type
    template[f"{path}/response_time"] = time
    if gain is not None:
        template[f"{path}/gain"] = gain

    if slit == "servo":
        template[f"{path}/slit/type"] = "servo"
    else:
        template[f"{path}/slit/type"] = "fixed"
        template[f"{path}/slit/x_gap"] = float(slit)
        template[f"{path}/slit/x_gap/@units"] = "nm"

    template[f"{path}/wavelength_range"] = wavelength_range

    return template


def read_detectors(metadata: list) -> Dict[str, Any]:
    """Reads detector values from the metadata and writes them into a template
    with the appropriate NeXus path."""

    template: Dict[str, Any] = {}
    detector_slits = parse_detector_line(metadata[31])
    detector_times = parse_detector_line(metadata[32], float)
    detector_gains = parse_detector_line(metadata[35], float)
    detector_changes = [float(x) for x in metadata[43].split()]
    wavelength_ranges = (
        [mpars.MIN_WAVELENGTH] + detector_changes[::-1] + [mpars.MAX_WAVELENGTH]
    )

    template.update(
        convert_detector_to_template(
            "PMT",
            detector_slits[-1],
            detector_times[-1],
            None,
            2,
            [wavelength_ranges[0], wavelength_ranges[1]],
        )
    )

    for name, idx, slit, time, gain in zip(
        ["PbS", "InGaAs"],
        [1, 0],
        detector_slits[1:],
        detector_times[1:],
        detector_gains[1:],
    ):
        template.update(
            convert_detector_to_template(
                name,
                slit,
                time,
                gain,
                idx,
                [wavelength_ranges[2 - idx], wavelength_ranges[3 - idx]],
            )
        )

    return template


def parse_asc(file_path: str) -> Dict[str, Any]:
    """Parses a Perkin Ellmer asc file into metadata and data dictionary.

    Args:
        file_path (str): File path to the asc file.

    Returns:
        Dict[str, Any]: Dictionary containing the metadata and data from the asc file.
    """
    template: Dict[str, Any] = {}
    data_start_ind = "#DATA"

    with open(file_path, encoding="utf-8") as fobj:
        metadata = []
        for line in fobj:
            if line.strip() == data_start_ind:
                break
            metadata.append(line.strip())

        data = pd.read_csv(fobj, delim_whitespace=True, header=None, index_col=0)

    for path, val in METADATA_MAP.items():
        # If the dict value is an int just get the data with it's index
        if isinstance(val, int):
            template[path] = metadata[val]
        elif isinstance(val, str):
            template[path] = val
        elif isfunction(val):
            template[path] = val(metadata)
        else:
            print(
                f"WARNING: "
                f"Invalid type value {type(val)} of entry '{path}:{val}' in METADATA_MAP"
            )

    template.update(read_detectors(metadata))
    template.update(data_to_template(data))

    return template


def add_def_info() -> Dict[str, str]:
    """Creates a template with definition version information"""
    template: Dict[str, Any] = {}
    template["/@default"] = "entry"
    template["/ENTRY[entry]/@default"] = "data"
    template["/ENTRY[entry]/definition"] = "NXtransmission"
    template["/ENTRY[entry]/definition/@version"] = "v2022.06"
    template["/ENTRY[entry]/definition/@url"] = (
        "https://fairmat-experimental.github.io/nexus-fairmat-proposal/"
        + "50433d9039b3f33299bab338998acb5335cd8951/index.html"
    )

    return template


# pylint: disable=too-few-public-methods
class TransmissionReader(YamlJsonReader):
    """TransmissionReader implementation for the DataConverter
    to convert transmission data to Nexus."""

    supported_nxdls = ["NXtransmission"]
    extensions = {
        ".asc": parse_asc,
        ".json": parse_json,
        ".yml": lambda fname: parse_yml(fname, CONVERT_DICT, REPLACE_NESTED),
        ".yaml": lambda fname: parse_yml(fname, CONVERT_DICT, REPLACE_NESTED),
        "default": lambda _: add_def_info(),
        "objects": handle_objects,
    }


READER = TransmissionReader



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""An example reader implementation for the DataConverter."""
import os
from typing import Tuple, Any
import math
from importlib.metadata import version
import yaml
import pandas as pd
import numpy as np
from pynxtools.dataconverter.readers.base.reader import BaseReader
from pynxtools.dataconverter.readers.ellips.mock import MockEllips
from pynxtools.dataconverter.helpers import extract_atom_types
from pynxtools.dataconverter.readers.utils import flatten_and_replace, FlattenSettings
from pynxtools import get_nexus_version, get_nexus_version_hash

DEFAULT_HEADER = {'sep': '\t', 'skip': 0}


CONVERT_DICT = {
    'unit': '@units',
    'Beam_path': 'BEAM_PATH[beam_path]',
    'Detector': 'DETECTOR[detector]',
    'Data': 'data_collection',
    'Derived_parameters': 'derived_parameters',
    'Environment': 'environment_conditions',
    'Instrument': 'INSTRUMENT[instrument]',
    'Sample': 'SAMPLE[sample]',
    'Sample_stage': 'sample_stage',
    'User': 'USER[user]',
    'Instrument/angle_of_incidence': 'INSTRUMENT[instrument]/angle_of_incidence',
    'Instrument/angle_of_incidence/unit': 'INSTRUMENT[instrument]/angle_of_incidence/@units',
    'column_names': 'data_collection/column_names',
    'data_error': 'data_collection/data_error',
    'depolarization': 'derived_parameters/depolarization',
    'measured_data': 'data_collection/measured_data',
    'software': 'software/program',
    'data_software': 'data_software/program',
}

CONFIG_KEYS = [
    'colnames',
    'derived_parameter_type',
    'err-var',
    'filename',
    'parameters',
    'plot_name',
    'sep',
    'skip',
    'spectrum_type',
    'spectrum_unit'
]

REPLACE_NESTED = {
    'Instrument/Beam_path': 'INSTRUMENT[instrument]/BEAM_PATH[beam_path]',
    'Env_Conditions': 'INSTRUMENT[instrument]/sample_stage/environment_conditions',
    'Instrument': 'INSTRUMENT[instrument]'
}


def load_header(filename, default):
    """ load the yaml description file, and apply defaults from
        the defalut dict for all keys not found from the file.

        Parameters:
            filename:           a yaml file containing the definitions
            default_header:     predefined default values

        Returns:
            a dict containing the loaded information
    """

    with open(filename, 'rt', encoding='utf8') as file:
        header = yaml.safe_load(file)

    clean_header = header

    for key, value in default.items():
        if key not in clean_header:
            clean_header[key] = value

    if 'sep' in header:
        clean_header['sep'] = header['sep'].encode("utf-8").decode("unicode_escape")

    return clean_header


def load_as_pandas_array(my_file, header):
    """ Load a CSV output file using the header dict.
        Use the fields: colnames, skip and sep from the header
        to instruct the csv reader about:
        colnames    -- column names
        skip        -- how many lines to skip
        sep         -- separator character in the file

        Parameters:
            my_file  string, file name
            header   dict header read from a yaml file

        Returns:
            A pandas array is returned.
    """
    required_parameters = ("colnames", "skip", "sep")
    for required_parameter in required_parameters:
        if required_parameter not in header:
            raise ValueError('colnames, skip and sep are required header parameters!')

    if not os.path.isfile(my_file):
        raise IOError(f'File not found error: {my_file}')

    whole_data = pd.read_csv(my_file,
                             # use header = None and names to define custom column names
                             header=None,
                             names=header['colnames'],
                             skiprows=header['skip'],
                             delimiter=header['sep'])
    return whole_data


def populate_header_dict(file_paths):
    """ This function creates and populates the header dictionary
        reading one or more yaml file.

        Parameters:
            file_paths  a list of file paths to be read

        Returns:
            a dict merging the content of all files
    """

    header = DEFAULT_HEADER

    for file_path in file_paths:
        if os.path.splitext(file_path)[1].lower() in [".yaml", ".yml"]:
            header = load_header(file_path, header)
            if "filename" not in header:
                raise KeyError("filename is missing from", file_path)
            data_file = os.path.join(os.path.split(file_path)[0], header["filename"])

            # if the path is not right, try the path provided directly
            if not os.path.isfile(data_file):
                data_file = header["filename"]

    return header, data_file


def populate_template_dict(header, template):
    """The template dictionary is then populated according to the content of header dictionary.

    """

    if "calibration_filename" in header:
        calibration = load_as_pandas_array(header["calibration_filename"], header)
        for k in calibration:
            header[f"calibration_{k}"] = calibration[k]

    eln_data_dict = flatten_and_replace(
        FlattenSettings(
            dic=header,
            convert_dict=CONVERT_DICT,
            replace_nested=REPLACE_NESTED,
            ignore_keys=CONFIG_KEYS
        )
    )
    template.update(eln_data_dict)

    return template


def header_labels(header, unique_angles):
    """ Define data labels (column names)

    """

    if header["Data"]["data_type"] == "Psi/Delta":
        labels = {"Psi": [], "Delta": []}
    elif header["Data"]["data_type"] == "tan(Psi)/cos(Delta)":
        labels = {"tan(Psi)": [], "cos(Delta)": []}
    else:
        labels = {}
        for i in range(1, 5):
            for j in range(1, 5):
                labels.update({f"m{i}{j}": []})

    for angle in enumerate(unique_angles):
        for key, val in labels.items():
            val.append(f"{key}_{int(angle[1])}deg")

    return labels


def mock_function(header):
    """ Mock ellipsometry data

    """

    mock_header = MockEllips(header)
    mock_header.mock_template(header)

    # Defining labels:
    mock_angles = header["Instrument/angle_of_incidence"]

    labels = header_labels(header, mock_angles)

    for angle in enumerate(header["Instrument/angle_of_incidence"]):
        for key, val in labels.items():
            val.append(f"{key}_{int(angle[1])}deg")

    return header, labels


def data_set_dims(whole_data):
    """ User defined variables to produce slices of the whole data set

    """
    energy = whole_data['type'].astype(str).values.tolist().count("E")
    unique_angles, counts = np.unique(whole_data["angle_of_incidence"
                                                 ].to_numpy()[0:energy].astype("int64"),
                                      return_counts=True
                                      )

    return unique_angles, counts


def parameter_array(whole_data, header, unique_angles, counts):
    """ User defined variables to produce slices of the whole data set

    """
    my_data_array = np.empty([
        len(unique_angles),
        1,
        counts[0]
    ])

    block_idx = [np.int64(0)]
    index = 0
    while index < len(whole_data):
        index += counts[0]
        block_idx.append(index)

    # derived parameters:
    # takes last but one column from the right (skips empty columns):
    data_index = 1
    temp = whole_data[header["colnames"][-data_index]].to_numpy()[
        block_idx[-1] - 1].astype("float64")

    while math.isnan(temp):
        temp = whole_data[header["colnames"][-data_index]].to_numpy()[
            block_idx[-1] - 1].astype("float64")
        data_index += 1

    for index in range(len(unique_angles)):
        my_data_array[
            index,
            0,
            :] = whole_data[header["colnames"][-data_index]].to_numpy()[
                block_idx[index + 6]:block_idx[index + 7]].astype("float64")

    return my_data_array


def data_array(whole_data, unique_angles, counts, labels):
    """ User defined variables to produce slices of the whole data set

    """
    my_data_array = np.empty([
        len(unique_angles),
        len(labels),
        counts[0]
    ])
    my_error_array = np.empty([
        len(unique_angles),
        len(labels),
        counts[0]
    ])

    block_idx = [np.int64(0)]
    index = 0
    while index < len(whole_data):
        index += counts[0]
        block_idx.append(index)

    data_index = 0
    for key, val in labels.items():
        for index in range(len(val)):
            my_data_array[
                index,
                data_index,
                :] = whole_data[key].to_numpy()[block_idx[index]:block_idx[index + 1]
                                                ].astype("float64")
        data_index += 1

    data_index = 0
    for key, val in labels.items():
        for index in range(len(val)):
            my_error_array[
                index,
                data_index,
                :] = whole_data[f"err.{key}"].to_numpy()[block_idx[index]:block_idx[index + 1]
                                                         ].astype("float64")
        data_index += 1

    return my_data_array, my_error_array


class EllipsometryReader(BaseReader):
    """An example reader implementation for the DataConverter.
    Importing metadata from the yaml file based on the last
    two parts of the key in the application definition.
    """

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = ["NXellipsometry"]

    @staticmethod
    def populate_header_dict_with_datasets(file_paths, is_mock=False):
        """This is an ellipsometry-specific processing of data.

        The procedure is the following:
        - the header dictionary is initialized reading a yaml file
        - the data are read from header["filename"] and saved in a pandas object
        - an array is shaped according to application definition in a 5D array (numpy array)
        - the array is saved in a HDF5 file as a dataset
        - virtual datasets instances are created in the header dictionary,
        referencing to data in the created HDF5 file.
        - the header is finally returned, ready to be parsed in the final template dictionary

        """
        header, data_file = populate_header_dict(file_paths)

        if os.path.isfile(data_file):
            whole_data = load_as_pandas_array(data_file, header)
        else:
            # this we have tried, we should throw an error...
            whole_data = load_as_pandas_array(header["filename"], header)

        unique_angles, counts = data_set_dims(whole_data)

        labels = header_labels(header, unique_angles)

        header["measured_data"], header["data_error"] = \
            data_array(whole_data, unique_angles, counts, labels)
        header[header["derived_parameter_type"]] = \
            parameter_array(whole_data, header, unique_angles, counts)

        spectrum_type = header["Data"]["spectrum_type"]
        if spectrum_type not in header["colnames"]:
            print("ERROR: spectrum type not found in 'colnames'")
        header[f"data_collection/NAME_spectrum[{spectrum_type}_spectrum]"] = (
            whole_data[spectrum_type].to_numpy()[0:counts[0]].astype("float64"))

        def write_scan_axis(name: str, values: list, units: str):
            base_path = f"Env_Conditions/PARAMETER[{name}]"
            header[f"{base_path}/values"] = values
            header[f"{base_path}/values/@units"] = units
            header[f"{base_path}/number_of_parameters"] = len(values)
            header[f"{base_path}/number_of_parameters/@units"] = ""
            header[f"{base_path}/parameter_type"] = name

        header["Instrument/angle_of_incidence"] = unique_angles
        for axis in ["detection_angle", "incident_angle"]:
            write_scan_axis(axis, unique_angles, "degree")

        # Create mocked ellipsometry data template:
        if is_mock:
            header, labels = mock_function(header)

        if "atom_types" not in header["Sample"]:
            header["atom_types"] = extract_atom_types(header["Sample"]["chemical_formula"])

        return header, labels

    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None,
             is_mock: bool = False) -> dict:
        """ Reads data from given file and returns a filled template dictionary.

            A handlings of virtual datasets is implemented:

            virtual dataset are created inside the final NeXus file.

            The template entry is filled with a dictionary containing the following keys:
            - link: the path of the external data file and the path of desired dataset inside it
            - shape: numpy array slice object (according to array slice notation)
        """

        if not file_paths:
            raise IOError("No input files were given to Ellipsometry Reader.")

        # The header dictionary is filled with entries.
        header, labels = (
            EllipsometryReader.populate_header_dict_with_datasets(file_paths, is_mock)
        )

        data_list = []
        for val in labels.values():
            data_list.append(val)

        # The template dictionary is filled
        template = populate_template_dict(header, template)

        spectrum_type = header["Data"]["spectrum_type"]
        if header["Data"]["spectrum_unit"] == "Angstroms":
            spectrum_unit = "angstrom"
        else:
            spectrum_unit = header["Data"]["spectrum_unit"]
        # MK:: Carola, Ron, Flo, Tamas, Sandor refactor the above-mentioned construct
        # there has to be a unit parsing control logic already at the level of this reader
        # because test-data.data has improper units like Angstroms or degrees
        # the fix above prevents that these incorrect units are get just blindly carried
        # over into the nxs file and thus causing nomas to fail
        template[f"/ENTRY[entry]/plot/AXISNAME[{spectrum_type}]"] = \
            {"link": f"/entry/data_collection/{spectrum_type}_spectrum"}
        template[f"/ENTRY[entry]/data_collection/NAME_spectrum[{spectrum_type}_spectrum]/@units"] \
            = spectrum_unit
        template[
            f"/ENTRY[entry]/data_collection/NAME_spectrum[{spectrum_type}_spectrum]/@long_name"
        ] = f"{spectrum_type} ({spectrum_unit})"
        plot_name = header["plot_name"]
        for dindx in range(0, len(labels.keys())):
            for index, key in enumerate(data_list[dindx]):
                template[f"/ENTRY[entry]/plot/DATA[{key}]"] = \
                    {
                        "link": "/entry/data_collection/measured_data",
                        "shape": np.index_exp[index, dindx, :]
                }
                # MK:: Carola, Ron, Flo, Tamas, Sandor refactor the following line
                # using a proper unit parsing logic
                template[f"/ENTRY[entry]/plot/DATA[{key}]/@units"] = "degree"
                if dindx == 0 and index == 0:
                    template[f"/ENTRY[entry]/plot/DATA[{key}]/@long_name"] = \
                        f"{plot_name} (degree)"
                template[f"/ENTRY[entry]/plot/DATA[{key}_errors]"] = \
                    {
                        "link": "/entry/data_collection/data_error",
                        "shape": np.index_exp[index, dindx, :]
                }
                # MK:: Carola, Ron, Flo, Tamas, Sandor refactor the following line
                template[f"/ENTRY[entry]/plot/DATA[{key}_errors]/@units"] = "degree"

        # Define default plot showing Psi and Delta at all angles:
        template["/@default"] = "entry"
        template["/ENTRY[entry]/@default"] = "plot"
        template["/ENTRY[entry]/plot/@signal"] = f"{data_list[0][0]}"
        template["/ENTRY[entry]/plot/@axes"] = spectrum_type
        template["/ENTRY[entry]/plot/title"] = plot_name

        # if len(data_list[0]) > 1:
        template["/ENTRY[entry]/plot/@auxiliary_signals"] = data_list[0][1:]
        for index in range(1, len(data_list)):
            template["/ENTRY[entry]/plot/@auxiliary_signals"] += data_list[index]

        template["/ENTRY[entry]/definition"] = "NXellipsometry"
        template["/ENTRY[entry]/definition/@url"] = (
            "https://github.com/FAIRmat-NFDI/nexus_definitions/"
            f"blob/{get_nexus_version_hash()}/contributed_definitions/NXellipsometry.nxdl.xml"
        )
        template["/ENTRY[entry]/definition/@version"] = get_nexus_version()
        template["/ENTRY[entry]/program_name"] = "pynxtools"
        template["/ENTRY[entry]/program_name/@version"] = version("pynxtools")
        template["/ENTRY[entry]/program_name/@url"] = "https://github.com/FAIRmat-NFDI/pynxtools"

        return template


# This has to be set to allow the convert script to use this reader. Set it to "MyDataReader".
READER = EllipsometryReader




# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Lake Shore Hall file reader implementation for the DataConverter."""
from pathlib import Path
import re
from typing import Any, List, TextIO, Dict, Optional
import logging
import numpy as np
import pandas as pd

from pynxtools.dataconverter.readers.json_yml.reader import YamlJsonReader
from pynxtools.dataconverter.readers.hall import helpers
from pynxtools.dataconverter.readers.utils import parse_json, parse_yml

# Replacement dict for section names
SECTION_REPLACEMENTS = {
    "Sample parameters": "entry/sample",
    "Measurements": "entry/measurement",
}

# Replacement dict for measurement indicators
MEASUREMENT_REPLACEMENTS = {
    "IV Curve Measurement": "iv_curve",
    "Variable Field Measurement": "variable_field",
}

# Dict for converting values for specific keys
CONVERSION_FUNCTIONS = {
    "Start Time": helpers.convert_date,
    "Time Completed": helpers.convert_date,
    "Skipped at": helpers.convert_date
}

# Keys that indicate the start of measurement block
MEASUREMENT_KEYS = ["Contact Sets"]

reader_dir = Path(__file__).parent
config_file = reader_dir.joinpath("enum_map.json")
ENUM_FIELDS = parse_json(str(config_file))

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)


def split_add_key(fobj: Optional[TextIO], dic: dict, prefix: str, expr: str) -> None:
    """Splits a key value pair and adds it to the dictionary.
    It also checks for measurement headers and adds the full tabular data as a
    pandas array to the dictionary.

    Args:
        fobj (TextIO): The file object to read from
        dic (dict): The dict to write the data into
        prefix (str): Key prefix for the dict
        expr (str): The current expr/line to parse
    """
    key, *val = re.split(r"\s*[:|=]\s*", expr)
    jval = "".join(val).strip()

    def parse_enum() -> bool:
        sprefix = prefix.strip("/")
        if 'Keithley' not in sprefix:
            w_trailing_num = re.search(r"(.*) \d+$", sprefix)
            if w_trailing_num:
                sprefix = w_trailing_num.group(1)

        if (
            sprefix in ENUM_FIELDS
            and key in ENUM_FIELDS[sprefix]
            and helpers.is_integer(jval)
        ):
            if jval not in ENUM_FIELDS[sprefix][key]:
                logger.warning("Option `%s` not in `%s, %s`", jval, sprefix, key)
            dic[f"{prefix}/{key}"] = ENUM_FIELDS[sprefix][key].get(jval, "UNKNOWN")
            return True

        return False

    def parse_field():
        if helpers.is_value_with_unit(jval):
            value, unit = helpers.split_value_with_unit(jval)
            dic[f"{prefix}/{key}"] = value
            dic[f"{prefix}/{key}/@units"] = helpers.clean(unit)
            return

        if parse_enum():
            return

        if helpers.is_integer(jval):
            dic[f"{prefix}/{key}"] = int(jval)
            return

        if helpers.is_number(jval):
            dic[f"{prefix}/{key}"] = float(jval)
            return

        if helpers.is_boolean(jval):
            dic[f"{prefix}/{key}"] = helpers.to_bool(jval)
            return

        dic[f"{prefix}/{key}"] = CONVERSION_FUNCTIONS.get(key, lambda v: v)(jval)

    def parse_data():
        data = []
        for line in fobj:
            if not line.strip():
                break
            if helpers.is_key(line):
                split_add_key(
                    None,  # There should be no deeper measurement,
                    # prevent further consumption of lines
                    dic,
                    f"{prefix}/{key}/{jval}",
                    line,
                )
                continue
            data.append(list(map(lambda x: x.strip(), re.split("\t+", line))))

        dkey = helpers.get_unique_dkey(dic, f"{prefix}/{key}/{jval}/data")
        dic[dkey] = pd.DataFrame(np.array(data[1:], dtype=np.float64), columns=data[0])

    if fobj is not None and key in MEASUREMENT_KEYS:
        parse_data()
    else:
        parse_field()


def parse_txt(fname: str, encoding: str = "iso-8859-1") -> dict:
    """Reads a template dictonary from a hall measurement file

    Args:
        fname (str): The file name of the masurement file
        encoding (str, optional): The encoding of the ASCII file. Defaults to "iso-8859-1".

    Returns:
        dict: Dict containing the data and metadata of the measurement
    """
    def parse_measurement(line: str, current_section: str, current_measurement: str):
        data = []
        for mline in fobj:
            if not mline.strip():
                break
            data.append(list(map(lambda x: x.strip(), re.split("\t+", mline))))

        header = list(map(lambda x: x.strip(), re.split("\t+", line)))
        dkey = helpers.get_unique_dkey(
            template, f"{current_section}{current_measurement}/data"
        )
        template.update(helpers.pandas_df_to_template(
            dkey,
            pd.DataFrame(
                np.array(data, dtype=np.float64), columns=header
            )
        ))

        return current_section, current_measurement

    def parse(line: str, current_section: str, current_measurement: str):
        if helpers.is_section(line):
            sline = line.strip()[1:-1]
            current_section = f"/{SECTION_REPLACEMENTS.get(sline, sline)}"
            current_measurement = ""
            return current_section, current_measurement

        if helpers.is_measurement(line):
            step, _, *meas = line.partition(":")
            sline = f"{step[6:]}_" + "".join(meas).strip()[:-1]
            current_measurement = f"/{MEASUREMENT_REPLACEMENTS.get(sline, sline)}"
            return current_section, current_measurement

        if helpers.is_key(line):
            split_add_key(
                fobj, template, f"{current_section}{current_measurement}", line
            )
            return current_section, current_measurement

        if helpers.is_meas_header(line):
            return parse_measurement(line, current_section, current_measurement)

        if line.strip():
            logger.warning("Line `%s` ignored", line.strip())

        return current_section, current_measurement

    template: Dict[str, Any] = {}
    current_section = "/entry"
    current_measurement = ""
    with open(fname, encoding=encoding) as fobj:
        for line in fobj:
            current_section, current_measurement = parse(
                line, current_section, current_measurement
            )

    return template


# pylint: disable=too-few-public-methods
class HallReader(YamlJsonReader):
    """HallReader implementation for the DataConverter
    to convert Hall data to Nexus."""

    supported_nxdls: List[str] = ["NXroot"]
    extensions = {
        ".txt": lambda fname: parse_txt(fname, "iso-8859-1"),
        ".json": parse_json,
        ".yml": lambda fname: parse_yml(fname, None, None),
        ".yaml": lambda fname: parse_yml(fname, None, None),
    }


READER = HallReader



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse intermediate HDF5 files from MTex @EBSD classes and FAIRmat mtex2nexus script."""

# pylint: disable=no-member

import numpy as np

import h5py


class NxEmOmMtexEbsdParser:
    """Parse *.mtex EBSD data.

    """

    def __init__(self, file_name: str, entry_id: int):
        """Class wrapping reading HDF5 files formatted according NXem_ebsd from MTex."""
        self.file_name = file_name
        self.entry_id = entry_id

    def parse_roi_default_plot(self, template: dict) -> dict:
        """Parse data for the region-of-interest default plot."""
        print("Parse ROI default plot...")
        h5r = h5py.File(self.file_name, "r")
        # by construction from MTex entry always named entry1

        src = "/entry1/indexing/region_of_interest"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/region_of_interest"
        template[f"{trg}/descriptor"] = str(h5r[f"{src}/descriptor"][()].decode("utf-8"))

        src = "/entry1/indexing/region_of_interest/roi"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/region_of_interest/roi"
        if src not in h5r.keys():
            # must not happen, grp is required
            # print(f"WARNING {src} not found !")
            return template
        grp = h5r[src]
        # MTex HDF5 file uses formatting from matching that of NXem_ebsd
        template[f"{trg}/title"] = str("Region-of-interest overview image")
        template[f"{trg}/@signal"] = grp.attrs["signal"]
        template[f"{trg}/@axes"] = grp.attrs["axes"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = grp.attrs["axis_x_indices"]
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = grp.attrs["axis_y_indices"]

        src = "/entry1/indexing/region_of_interest/roi/data"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/region_of_interest/roi/data"
        if src not in h5r.keys():
            # must not happen, dst is required
            print(f"{src} not found !")
            return template
        dst = h5r[src]
        template[f"{trg}"] = {"compress": dst[:, :], "strength": 1}
        template[f"{trg}/@long_name"] = dst.attrs["long_name"]
        template[f"{trg}/@CLASS"] = "IMAGE"  # required by H5Web to plot RGB maps
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        axes_names = ["axis_x", "axis_y"]
        for axis_name in axes_names:
            src = f"/entry1/indexing/region_of_interest/roi/{axis_name}"
            trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
                  f"region_of_interest/roi/{axis_name}"
            if src not in h5r.keys():
                # must not happen, dst is required
                # print(f"{src} not found !")
                return template
            dst = h5r[src]
            template[f"{trg}"] = {"compress": dst[:], "strength": 1}
            template[f"{trg}/@long_name"] = dst.attrs["long_name"]
            template[f"{trg}/@units"] = dst.attrs["units"]

        h5r.close()
        return template

    def parse_phases(self, template: dict) -> dict:
        """Parse data for the crystal structure models aka phases."""
        print("Parse crystal_structure_models aka phases...")
        h5r = h5py.File(self.file_name, "r")

        src = "/entry1/indexing"
        # mtex2nexus MTex/Matlab scripts writes controlled terms phaseID
        group_names = [entry for entry in h5r[src].keys() if entry.startswith('phase')]
        if len(group_names) == 0:
            return template
        # group_names end up sorted in ascending order
        identifier = 1
        for group_name in group_names:
            if f"phase{identifier}" != group_name:
                # must not happen, verifier will complain
                return template
            src = f"/entry1/indexing/{group_name}"
            trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
                  f"EM_EBSD_CRYSTAL_STRUCTURE_MODEL" \
                  f"[em_ebsd_crystal_structure_model{identifier}]"
            template[f"{trg}/phase_identifier"] = h5r[f"{src}/phase_identifier"][0]
            template[f"{trg}/phase_name"] \
                = str(h5r[f"{src}/phase_name"][()].decode("utf-8"))
            template[f"{trg}/point_group"] \
                = str(h5r[f"{src}/point_group"][()].decode("utf-8"))
            dst = h5r[f"{src}/unit_cell_abc"]
            template[f"{trg}/unit_cell_abc"] = dst[:]
            template[f"{trg}/unit_cell_abc/@units"] = dst.attrs["units"]
            dst = h5r[f"{src}/unit_cell_alphabetagamma"]
            template[f"{trg}/unit_cell_alphabetagamma"] = dst[:]
            template[f"{trg}/unit_cell_alphabetagamma/@units"] = dst.attrs["units"]
            identifier += 1
        h5r.close()
        return template

    def parse_inverse_pole_figures(self, template: dict) -> dict:
        """Parse inverse pole figures (IPF) mappings."""
        print("Parse inverse pole figures (IPFs)...")
        h5r = h5py.File(self.file_name, "r")

        src = "/entry1/indexing"
        # mtex2nexus MTex/Matlab scripts writes controlled terms phaseID
        group_names = [entry for entry in h5r[src].keys() if entry.startswith('ipf_map')]
        if len(group_names) == 0:
            return template
        # group_names end up sorted in ascending order
        identifier = 1
        for group_name in group_names:
            if f"ipf_map{identifier}" != group_name:
                # must not happen, verifier will complain that we use incorrect names
                # print(f"WARNING: {group_name} not found !")
                return template

            self.parse_inverse_pole_figure_map(h5r, identifier, template)
            self.parse_inverse_pole_figure_color_key(h5r, identifier, template)

            identifier += 1

        h5r.close()
        return template

    def parse_inverse_pole_figure_map(self, h5r, identifier, template: dict) -> dict:
        """Parse inverse-pole-figure (IPF) mappings on their color models."""
        group_name = f"ipf_map{identifier}"
        print(f"Parse inverse pole figure (IPF) map for {group_name}...")
        src = f"/entry1/indexing/{group_name}"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/PROCESS[ipf_map{identifier}]"
        if src not in h5r.keys():
            # print(f"WARNING: {group_name} not found !")
            return template
        template[f"{trg}/bitdepth"] = np.uint32(8)  # h5r[f"{src}/bitdepth"][0]
        template[f"{trg}/phase_identifier"] = np.uint32(h5r[f"{src}/phase_identifier"][0])
        template[f"{trg}/phase_name"] = str(h5r[f"{src}/phase_name"][()].decode("utf-8"))
        dst = h5r[f"{src}/program"]
        template[f"{trg}/PROGRAM[program1]/program"] = str(dst[()].decode("utf-8"))
        template[f"{trg}/PROGRAM[program1]/program/@version"] = dst.attrs["version"]
        template[f"{trg}/projection_direction"] = np.asarray([0., 0., 1.], np.float32)
        # there should be a depends on etc

        src = f"/entry1/indexing/{group_name}/ipf_rgb_map"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
              f"PROCESS[ipf_map{identifier}]/ipf_rgb_map"
        if src not in h5r.keys():
            # must not happen, grp is required
            # print(f"WARNING: {group_name} not found, ipf_rgb_map !")
            return template
        grp = h5r[src]
        # MTex HDF5 file uses formatting from matching that of NXem_ebsd
        template[f"{trg}/title"] = str("Inverse pole figure color map")
        template[f"{trg}/@signal"] = grp.attrs["signal"]
        template[f"{trg}/@axes"] = grp.attrs["axes"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = grp.attrs["axis_x_indices"]
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = grp.attrs["axis_y_indices"]

        src = f"/entry1/indexing/{group_name}/ipf_rgb_map/data"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
              f"PROCESS[ipf_map{identifier}]/ipf_rgb_map/DATA[data]"
        if src not in h5r.keys():
            # must not happen, dst is required
            # print(f"WARNING: {group_name} not found, ipf_rgb_map, data !")
            return template
        dst = h5r[src]
        template[f"{trg}"] = {"compress": dst[:, :], "strength": 1}
        template[f"{trg}/@long_name"] = dst.attrs["long_name"]
        template[f"{trg}/@CLASS"] = "IMAGE"  # required by H5Web to plot RGB maps
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        axes_names = ["axis_x", "axis_y"]
        for axis_name in axes_names:
            src = f"/entry1/indexing/{group_name}/ipf_rgb_map/{axis_name}"
            trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
                  f"/PROCESS[ipf_map{identifier}]/ipf_rgb_map/AXISNAME[{axis_name}]"
            if src not in h5r.keys():
                # must not happen, dst is required
                # print(f"WARNING: {group_name} not found, ipf_rgb_map, {axis_name} !")
                return template
            dst = h5r[src]
            template[f"{trg}"] = {"compress": dst[:], "strength": 1}
            template[f"{trg}/@long_name"] = dst.attrs["long_name"]
            template[f"{trg}/@units"] = dst.attrs["units"]

        return template

    def parse_inverse_pole_figure_color_key(self, h5r, identifier, template: dict) -> dict:
        """Parse color key renderings of inverse-pole-figure (IPF) mappings."""
        group_name = f"ipf_map{identifier}"
        print(f"Parse inverse pole figure (IPF) color key {group_name}...")
        src = f"/entry1/indexing/{group_name}/ipf_rgb_color_model"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/PROCESS[ipf_map{identifier}]/ipf_rgb_color_model"
        if src not in h5r.keys():
            # must not happen, grp is required
            # print(f"WARNING: {group_name} not found, ipf_rgb_color_model")
            return template
        grp = h5r[src]
        template[f"{trg}/title"] = str("Inverse pole figure color key with SST")
        template[f"{trg}/@signal"] = grp.attrs["signal"]
        template[f"{trg}/@axes"] = grp.attrs["axes"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = grp.attrs["axis_x_indices"]
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = grp.attrs["axis_y_indices"]

        src = f"/entry1/indexing/{group_name}/ipf_rgb_color_model/data"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/PROCESS[ipf_map{identifier}]/ipf_rgb_color_model/DATA[data]"
        if src not in h5r.keys():
            # must not happen, dst is required
            # print(f"WARNING: {group_name} not found, ipf_rgb_color_model, data")
            return template
        dst = h5r[src]
        template[f"{trg}"] = {"compress": dst[:, :], "strength": 1}
        template[f"{trg}/@long_name"] = dst.attrs["long_name"]
        template[f"{trg}/@CLASS"] = "IMAGE"
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        axes_names = ["axis_x", "axis_y"]
        for axis_name in axes_names:
            src = f"/entry1/indexing/{group_name}/ipf_rgb_color_model/{axis_name}"
            trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
                  f"/PROCESS[ipf_map{identifier}]/ipf_rgb_color_model" \
                  f"/AXISNAME[{axis_name}]"
            if src not in h5r.keys():
                # must not happen, dst is required
                # print(f"WARNING: {group_name} not found,
                # ipf_rgb_color_model, {axis_name}")
                return template
            dst = h5r[src]
            template[f"{trg}"] = {"compress": dst[:], "strength": 1}
            template[f"{trg}/@long_name"] = dst.attrs["long_name"]
            template[f"{trg}/@units"] = str("px")

        return template

    def parse(self, template: dict) -> dict:
        """Parse metadata and numerical data."""
        print("Parsing MTex EBSD data...")
        print(f"{self.file_name}")
        print(f"{self.entry_id}")
        # extract data from intermediate HDF5 file and transfer these to the template
        self.parse_roi_default_plot(template)
        self.parse_phases(template)
        self.parse_inverse_pole_figures(template)
        return template



#!/usr/bin/env python3
"""Main file of nyaml2nxdl tool.
Users create NeXus instances by writing a YAML file
which details a hierarchy of data/metadata elements

"""
# -*- coding: utf-8 -*-
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import xml.etree.ElementTree as ET

import click

from .nyaml2nxdl_backward_tools import Nxdl2yaml
from .nyaml2nxdl_backward_tools import compare_niac_and_my
from .nyaml2nxdl_forward_tools import nyaml2nxdl
from .nyaml2nxdl_forward_tools import pretty_print_xml
from .nyaml2nxdl_helper import extend_yamlfile_by_nxdl_as_comment
from .nyaml2nxdl_helper import get_sha256_hash
from .nyaml2nxdl_helper import separate_hash_yaml_and_nxdl

DEPTH_SIZE = 4 * " "

# NOTE: Some handful links for nyaml2nxdl converter:
# https://manual.nexusformat.org/nxdl_desc.html?highlight=optional


def generate_nxdl_or_retrieve_nxdl(yaml_in, xml_out, verbose):
    """
    Generate yaml, nxdl, and hash.
        If the extracted hash is exactly the same as generated from input yaml then
        retrieve the nxdl part from provided yaml and return nxdl as output.
        Else, generate nxdl from input yaml with the help of nyaml2nxdl function
    """

    file_path, rel_file = os.path.split(yaml_in)
    sep_yaml = os.path.join(file_path, f"temp_{rel_file}")
    hash_found = separate_hash_yaml_and_nxdl(yaml_in, sep_yaml, xml_out)

    if hash_found:
        gen_hash = get_sha256_hash(sep_yaml)
        if hash_found == gen_hash:
            os.remove(sep_yaml)
            return

    nyaml2nxdl(sep_yaml, xml_out, verbose)
    os.remove(sep_yaml)


# pylint: disable=too-many-locals
def append_yml(input_file, append, verbose):
    """Append to an existing NeXus base class new elements provided in YML input file \
and print both an XML and YML file of the extended base class.

"""
    nexus_def_path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../../definitions"
    )
    assert [
        s
        for s in os.listdir(os.path.join(nexus_def_path, "base_classes"))
        if append.strip() == s.replace(".nxdl.xml", "")
    ], "Your base class extension does not match any existing NeXus base classes"
    tree = ET.parse(
        os.path.join(nexus_def_path + "/base_classes", append + ".nxdl.xml")
    )
    root = tree.getroot()
    # warning: tmp files are printed on disk and removed at the ends!!
    pretty_print_xml(root, "tmp.nxdl.xml")
    input_tmp_xml = "tmp.nxdl.xml"
    out_tmp_yml = "tmp_parsed.yaml"
    converter = Nxdl2yaml([], [])
    converter.print_yml(input_tmp_xml, out_tmp_yml, verbose)
    nyaml2nxdl(input_file=out_tmp_yml, out_file="tmp_parsed.nxdl.xml", verbose=verbose)
    tree = ET.parse("tmp_parsed.nxdl.xml")
    tree2 = ET.parse(input_file)
    root_no_duplicates = ET.Element(
        "definition",
        {
            "xmlns": "http://definition.nexusformat.org/nxdl/3.1",
            "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "xsi:schemaLocation": "http://www.w3.org/2001/XMLSchema-instance",
        },
    )
    for attribute_keys in root.attrib.keys():
        if (
            attribute_keys
            != "{http://www.w3.org/2001/XMLSchema-instance}schemaLocation"
        ):
            attribute_value = root.attrib[attribute_keys]
            root_no_duplicates.set(attribute_keys, attribute_value)
    for elems in root.iter():
        if "doc" in elems.tag:
            root_doc = ET.SubElement(root_no_duplicates, "doc")
            root_doc.text = elems.text
            break
    group = "{http://definition.nexusformat.org/nxdl/3.1}group"
    root_no_duplicates = compare_niac_and_my(
        tree, tree2, verbose, group, root_no_duplicates
    )
    field = "{http://definition.nexusformat.org/nxdl/3.1}field"
    root_no_duplicates = compare_niac_and_my(
        tree, tree2, verbose, field, root_no_duplicates
    )
    attribute = "{http://definition.nexusformat.org/nxdl/3.1}attribute"
    root_no_duplicates = compare_niac_and_my(
        tree, tree2, verbose, attribute, root_no_duplicates
    )
    pretty_print_xml(
        root_no_duplicates,
        f"{input_file.replace('.nxdl.xml', '')}" f"_appended.nxdl.xml",
    )

    input_file_xml = input_file.replace(".nxdl.xml", "_appended.nxdl.xml")
    out_file_yml = input_file.replace(".nxdl.xml", "_appended_parsed.yaml")
    converter = Nxdl2yaml([], [])
    converter.print_yml(input_file_xml, out_file_yml, verbose)
    nyaml2nxdl(
        input_file=out_file_yml,
        out_file=out_file_yml.replace(".yaml", ".nxdl.xml"),
        verbose=verbose,
    )
    os.rename(
        f"{input_file.replace('.nxdl.xml', '_appended_parsed.yaml')}",
        f"{input_file.replace('.nxdl.xml', '_appended.yaml')}",
    )
    os.rename(
        f"{input_file.replace('.nxdl.xml', '_appended_parsed.nxdl.xml')}",
        f"{input_file.replace('.nxdl.xml', '_appended.nxdl.xml')}",
    )
    os.remove("tmp.nxdl.xml")
    os.remove("tmp_parsed.yaml")
    os.remove("tmp_parsed.nxdl.xml")


def split_name_and_extension(file_name):
    """
    Split file name into extension and rest of the file name.
    return file raw nam and extension
    """
    path = file_name.rsplit("/", 1)
    (pathn, filen) = ["", path[0]] if len(path) == 1 else [path[0] + "/", path[1]]
    parts = filen.rsplit(".", 2)
    raw = ext = ""
    if len(parts) == 2:
        raw = parts[0]
        ext = parts[1]
    elif len(parts) == 3:
        raw = parts[0]
        ext = ".".join(parts[1:])

    return pathn + raw, ext


@click.command()
@click.option(
    "--input-file",
    required=True,
    prompt=True,
    help="The path to the XML or YAML input data file to read and create \
a YAML or XML file from, respectively.",
)
@click.option(
    "--append",
    help="Parse xml file and append to base class, given that the xml file has same name \
of an existing base class",
)
@click.option(
    "--check-consistency",
    is_flag=True,
    default=False,
    help=(
        "Check wether yaml or nxdl has followed general rules of scema or not"
        "check whether your comment in the right place or not. The option render an "
        "output file of the same extension(*_consistency.yaml or *_consistency.nxdl.xml)"
    ),
)
@click.option(
    "--do-not-store-nxdl",
    is_flag=True,
    default=True,
    help=(
        "Whether the input nxdl file will be stored as a comment"
        " at the end of output yaml file."
    ),
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Print in standard output keywords and value types to help \
possible issues in yaml files",
)
def launch_tool(input_file, verbose, append, do_not_store_nxdl, check_consistency):
    """
    Main function that distiguishes the input file format and launches the tools.
    """
    if os.path.isfile(input_file):
        raw_name, ext = split_name_and_extension(input_file)
    else:
        raise ValueError("Need a valid input file.")

    if ext == "yaml":
        xml_out_file = raw_name + ".nxdl.xml"
        generate_nxdl_or_retrieve_nxdl(input_file, xml_out_file, verbose)
        if append:
            append_yml(raw_name + ".nxdl.xml", append, verbose)
        # For consistency running
        if check_consistency:
            yaml_out_file = raw_name + "_consistency." + ext
            converter = Nxdl2yaml([], [])
            converter.print_yml(xml_out_file, yaml_out_file, verbose)
            os.remove(xml_out_file)
    elif ext == "nxdl.xml":
        if not append:
            yaml_out_file = raw_name + "_parsed" + ".yaml"
            converter = Nxdl2yaml([], [])
            converter.print_yml(input_file, yaml_out_file, verbose)
            # Store nxdl.xml file in output yaml file under SHA HASH
            yaml_hash = get_sha256_hash(yaml_out_file)
            # Lines as divider between yaml and nxdl
            top_lines = [
                (
                    "\n# ++++++++++++++++++++++++++++++++++ SHA HASH"
                    " ++++++++++++++++++++++++++++++++++\n"
                ),
                f"# {yaml_hash}\n",
            ]
            if do_not_store_nxdl:
                extend_yamlfile_by_nxdl_as_comment(
                    yaml_file=yaml_out_file,
                    file_to_be_appended=input_file,
                    top_lines_list=top_lines,
                )
        else:
            append_yml(input_file, append, verbose)
        # Taking care of consistency running
        if check_consistency:
            xml_out_file = raw_name + "_consistency." + ext
            generate_nxdl_or_retrieve_nxdl(yaml_out_file, xml_out_file, verbose)
            os.remove(yaml_out_file)
    else:
        raise ValueError("Provide correct file with extension '.yaml or '.nxdl.xml")


if __name__ == "__main__":
    launch_tool().parse()  # pylint: disable=no-value-for-parameter



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility functions for the NeXus reader classes."""
import logging
from dataclasses import dataclass, replace
from typing import List, Any, Dict, Optional, Tuple
from collections.abc import Mapping
import json
import yaml

logger = logging.getLogger(__name__)


@dataclass
class FlattenSettings():
    """Settings for flattening operations.

    Args:
        dic (dict): Dictionary to flatten
        convert_dict (dict): Dictionary for renaming keys in the flattend dict.
        replace_nested (dict): Dictionary for renaming nested keys.
        parent_key (str, optional):
            Parent key of the dictionary. Defaults to "/ENTRY[entry]".
        sep (str, optional): Separator for the keys. Defaults to "/".
    """
    dic: Mapping
    convert_dict: dict
    replace_nested: dict
    parent_key: str = "/ENTRY[entry]"
    sep: str = "/"
    is_in_section: bool = False
    ignore_keys: Optional[list] = None


def is_section(val: Any) -> bool:
    """Checks whether a value is a section.

    Args:
        val (Any): A list or value.

    Returns:
        bool: True if val is a section.
    """
    return isinstance(val, list) and len(val) > 0 and isinstance(val[0], dict)


def is_value_unit_pair(val: Any) -> bool:
    """Checks whether the value contains a dict of a value unit pair.

    Args:
        val (Any): The value to be checked.

    Returns:
        bool: True if val contains a value unit pair dict.
    """
    if not isinstance(val, dict):
        return False

    if len(val) == 2 and "value" in val and "unit" in val:
        return True
    return False


def uniquify_keys(ldic: list) -> List[Any]:
    """Uniquifys keys in a list of tuple lists containing key value pairs.

    Args:
        ldic (list): List of lists of length two, containing key value pairs.

    Returns:
        List[Any]: Uniquified list, where duplicate keys are appended with 1, 2, etc.
    """
    dic: dict = {}
    for key, val in ldic:
        suffix = 0
        sstr = "" if suffix == 0 else str(suffix)
        while f"{key}{sstr}" in dic:
            sstr = "" if suffix == 0 else str(suffix)
            suffix += 1

        if is_value_unit_pair(val):
            dic[f"{key}{sstr}"] = val["value"]
            dic[f"{key}{sstr}/@units"] = val["unit"]
            continue
        dic[f"{key}{sstr}"] = val

    return list(map(list, dic.items()))


def parse_section(key: str, val: Any, settings: FlattenSettings) -> List[Any]:
    """Parse a section, i.e. an entry containing a list of entries.

    Args:
        key (str): The key which is currently being checked.
        val (Any): The value at the current key.
        settings (FlattenSettings): The flattening settings.

    Returns:
        List[Any]: A list of list tuples containing key, value pairs.
    """
    if not is_section(val):
        return [(key, val)]

    groups: List[Any] = []
    for group in val:
        groups.extend(
            flatten_and_replace(
                replace(settings, dic=group, parent_key=key, is_in_section=True)
            ).items()
        )

    return uniquify_keys(groups)


def flatten_and_replace(settings: FlattenSettings) -> dict:
    """Flatten a nested dictionary, and replace the keys with the appropriate
    paths in the nxs file.

    Args:
        settings (FlattenSettings): Settings dataclass for flattening the data.

    Returns:
        dict: Flattened dictionary
    """
    items: List[Any] = []
    for key, val in settings.dic.items():
        if settings.ignore_keys and key in settings.ignore_keys:
            continue
        new_key = settings.parent_key + settings.sep + settings.convert_dict.get(key, key)
        if isinstance(val, Mapping):
            items.extend(
                flatten_and_replace(replace(settings, dic=val, parent_key=new_key))
                .items()
                if not (settings.is_in_section and is_value_unit_pair(val))
                else [[new_key, val]]
            )
            continue

        for old, new in settings.replace_nested.items():
            new_key = new_key.replace(old, new)

        if new_key.endswith("/value"):
            items.append((new_key[:-6], val))
        else:
            items.extend(parse_section(new_key, val, settings))

    return dict(items)


def parse_yml(
        file_path: str,
        convert_dict: Optional[dict] = None,
        replace_nested: Optional[dict] = None
) -> Dict[str, Any]:
    """Parses a metadata yaml file into a dictionary.

    Args:
        file_path (str): The file path of the yml file.

    Returns:
        Dict[str, Any]: The dictionary containing the data readout from the yml.
    """
    if convert_dict is None:
        convert_dict = {}

    if replace_nested is None:
        replace_nested = {}

    convert_dict["unit"] = "@units"

    with open(file_path, encoding='utf-8') as file:
        return flatten_and_replace(
            FlattenSettings(
                dic=yaml.safe_load(file),
                convert_dict=convert_dict,
                replace_nested=replace_nested
            )
        )


def parse_json(file_path: str) -> Dict[str, Any]:
    """Parses a metadata json file into a dictionary.

    Args:
        file_path (str): The file path of the json file.

    Returns:
        Dict[str, Any]: The dictionary containing the data readout from the json.
    """
    with open(file_path, "r", encoding="utf-8") as file:
        return json.load(file)


def handle_objects(objects: Tuple[Any]) -> Dict[str, Any]:
    """Handle objects and generate template entries from them"""
    if objects is None:
        return {}

    template = {}

    for obj in objects:
        if not isinstance(obj, dict):
            logger.warning("Ignoring unknown object of type %s", type(obj))
            continue

        template.update(obj)

    return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Logic for annotating existent default plots in NXdata for NXem_ebsd for H5Web."""

# pylint: disable=no-member

HFIVE_WEB_MAX_SIZE = 2048


def roi_plot_available(template: dict, entry_id: int) -> bool:
    """Choose a preferred NXdata/data instance either the ROI or IPF map."""
    # most interesting to show are three-dimensional orientation mappings ...
    trg = f"/ENTRY[entry{entry_id}]/correlation/region_of_interest/roi/data"
    if trg in template.keys():
        print(f"Found image default plot for entry{entry_id}")
        trg = "/"
        template[f"{trg}@default"] = f"entry{entry_id}"
        trg += f"ENTRY[entry{entry_id}]/"
        template[f"{trg}@default"] = "correlation"
        trg += "correlation/"
        template[f"{trg}@default"] = "region_of_interest"
        trg += "region_of_interest/"
        template[f"{trg}@default"] = "roi"
        return True

    # fall-back display an IPF mapping
    trg = f"/ENTRY[entry{entry_id}]/correlation/PROCESS[ipf_map1]/ipf_rgb_map/data"
    # by definition if there is one it will be always the first one and we use this IPF map
    if trg in template.keys():
        print(f"Found image default plot for entry{entry_id}")
        trg = "/"
        template[f"{trg}@default"] = f"entry{entry_id}"
        trg += f"ENTRY[entry{entry_id}]/"
        template[f"{trg}@default"] = "correlation"
        trg += "correlation/"
        template[f"{trg}@default"] = "ipf_map1"
        trg += "PROCESS[ipf_map1]/"
        template[f"{trg}@default"] = "ipf_rgb_map"
        return True

    # ... but in most practical cases we have two-dimensional EBSD maps only...
    trg = f"/ENTRY[entry{entry_id}]/experiment/indexing/region_of_interest/roi/data"
    if trg in template.keys():
        print(f"Found image default plot for entry{entry_id}")
        trg = "/"
        template[f"{trg}@default"] = f"entry{entry_id}"
        trg += f"ENTRY[entry{entry_id}]/"
        template[f"{trg}@default"] = "experiment"
        trg += "experiment/"
        template[f"{trg}@default"] = "indexing"
        trg += "indexing/"
        template[f"{trg}@default"] = "region_of_interest"
        return True

    # fall-back display an IPF mapping
    trg = f"/ENTRY[entry{entry_id}]/experiment/indexing" \
          f"/PROCESS[ipf_map1]/ipf_rgb_map/data"
    # by definition if there is one it will be always the first one and we use this IPF map
    if trg in template.keys():
        print(f"Found image default plot for entry{entry_id}")
        trg = "/"
        template[f"{trg}@default"] = f"entry{entry_id}"
        trg += f"ENTRY[entry{entry_id}]/"
        template[f"{trg}@default"] = "experiment"
        trg += "experiment/"
        template[f"{trg}@default"] = "indexing"
        trg += "indexing/"
        template[f"{trg}@default"] = "ipf_map1"
        trg += "PROCESS[ipf_map1]/"
        template[f"{trg}@default"] = "ipf_rgb_map"
        return True

    return False


def diffraction_pattern_available(template: dict, entry_id: int) -> bool:
    """Choose a preferred NXdata/data instance diffraction pattern."""
    trg = f"/ENTRY[entry{entry_id}]/simulation/IMAGE_SET_EM_KIKUCHI" \
          f"[image_set_em_kikuchi]/stack/data_counts"
    if trg in template.keys():
        print(f"Found image default plot for entry{entry_id}")
        trg = "/"
        template[f"{trg}@default"] = f"entry{entry_id}"
        trg += f"ENTRY[entry{entry_id}]/"
        template[f"{trg}@default"] = "simulation"
        trg += "simulation/"
        template[f"{trg}@default"] = "image_set_em_kikuchi"
        trg += "image_set_em_kikuchi/"
        template[f"{trg}@default"] = "stack"
        return True

    return False


def em_om_default_plot_generator(template: dict, entry_id: int) -> dict:
    """Identify and annotate default plot."""
    # prefer to show the region-of-interest overview as the default plot
    # if this does not exist, try to find one of the IPF color mappings
    if roi_plot_available(template, entry_id) is True:
        return template

    if diffraction_pattern_available(template, entry_id) is True:
        return template

    print(f"WARNING: No path to a default plot found for entry{entry_id} !")
    return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse data from custom schema YAML file onto NeXus concept instances in NXem."""

# pylint: disable=no-member

import numpy as np

import flatdict as fd

import yaml

from ase.data import chemical_symbols

from pynxtools.dataconverter.readers.em_nion.utils.em_nion_versioning \
    import NX_EM_NION_ADEF_NAME, NX_EM_NION_ADEF_VERSION

from pynxtools.dataconverter.readers.em_nion.utils.em_nion_versioning \
    import NX_EM_NION_EXEC_NAME, NX_EM_NION_EXEC_VERSION

from pynxtools.dataconverter.readers.shared.map_concepts.mapping_functors \
    import apply_modifier, variadic_path_to_specific_path

from pynxtools.dataconverter.readers.em_nion.map_concepts.swift_eln_to_nx_map \
    import NxEmElnInput, NxUserFromListOfDict, NxDetectorListOfDict, NxSample


class NxEmNionElnSchemaParser:
    """Parse eln_data.yaml dump file content generated from a NOMAD OASIS YAML.

    The functionalities in this ELN YAML parser do not check if the
    instantiated template yields an instance which is compliant NXapm.
    Instead, this task is handled by the generic part of the dataconverter
    during the verification of the template dictionary.
    """

    def __init__(self, file_name: str, entry_id: int):
        print(f"Extracting data from ELN file: {file_name}")
        if (file_name.rsplit('/', 1)[-1].startswith("eln_data")
                or file_name.startswith("eln_data")) and entry_id > 0:
            self.entry_id = entry_id
            self.file_name = file_name
            with open(self.file_name, "r", encoding="utf-8") as stream:
                self.yml = fd.FlatDict(yaml.safe_load(stream), delimiter="/")
        else:
            self.entry_id = 1
            self.file_name = ""
            self.yml = {}

    def parse_user_section(self, template: dict) -> dict:
        """Copy data from user section into template."""
        src = "user"
        if src in self.yml.keys():
            if isinstance(self.yml[src], list):
                if (all(isinstance(entry, dict) for entry in self.yml[src]) is True):
                    user_id = 1
                    # custom schema delivers a list of dictionaries...
                    for user_dict in self.yml[src]:
                        # ... for each of them inspect for fields mappable on NeXus
                        identifier = [self.entry_id, user_id]
                        # identifier to get instance NeXus path from variadic NeXus path
                        # try to find all quantities on the left-hand side of the mapping
                        # table and check if we can find these
                        for nx_path, modifier in NxUserFromListOfDict.items():
                            if (nx_path != "IGNORE") and (nx_path != "UNCLEAR"):
                                trg = variadic_path_to_specific_path(nx_path, identifier)
                                res = apply_modifier(modifier, user_dict)
                                if res is not None:
                                    template[trg] = res
                        user_id += 1
        return template

    def parse_sample_section(self, template: dict) -> dict:
        """Copy data from sample section into template."""
        src = "sample/atom_types"
        trg = f"/ENTRY[entry{self.entry_id}]/{src}"
        if "sample/atom_types" in self.yml.keys():
            if (isinstance(self.yml[src], list)) and (len(self.yml[src]) >= 1):
                atom_types_are_valid = True
                for symbol in self.yml[src]:
                    valid = isinstance(symbol, str) \
                        and (symbol in chemical_symbols) and (symbol != "X")
                    if valid is False:
                        atom_types_are_valid = False
                        break
                if atom_types_are_valid is True:
                    template[trg] = ", ".join(list(self.yml[src]))

        for nx_path, modifier in NxSample.items():
            if (nx_path != "IGNORE") and (nx_path != "UNCLEAR"):
                trg = variadic_path_to_specific_path(nx_path, [self.entry_id])
                res = apply_modifier(modifier, self.yml)
                if res is not None:
                    template[trg] = res

        return template

    def parse_detector_section(self, template: dict) -> dict:
        """Copy data from detector section into template."""
        src = "em_lab/detector"
        if src in self.yml.keys():
            if isinstance(self.yml[src], list):
                if (all(isinstance(entry, dict) for entry in self.yml[src]) is True):
                    detector_id = 1
                    # custom schema delivers a list of dictionaries...
                    for detector_dict in self.yml[src]:
                        # ... for each of them inspect for fields mappable on NeXus
                        identifier = [self.entry_id, detector_id]
                        # identifier to get instance NeXus path from variadic NeXus path
                        # try to find all quantities on the left-hand side of the mapping
                        # table and check if we can find these
                        for nx_path, modifier in NxDetectorListOfDict.items():
                            if (nx_path != "IGNORE") and (nx_path != "UNCLEAR"):
                                trg = variadic_path_to_specific_path(nx_path, identifier)
                                res = apply_modifier(modifier, detector_dict)
                                if res is not None:
                                    template[trg] = res
                        detector_id += 1

        return template

    def parse_other_sections(self, template: dict) -> dict:
        """Copy data from custom schema (excluding user, sample) into template."""
        for nx_path, modifier in NxEmElnInput.items():
            if (nx_path != "IGNORE") and (nx_path != "UNCLEAR"):
                trg = variadic_path_to_specific_path(nx_path, [self.entry_id, 1])
                res = apply_modifier(modifier, self.yml)
                if res is not None:
                    template[trg] = res
        return template

    def parse(self, template: dict) -> dict:
        """Copy data from self into template the appdef instance."""
        self.parse_user_section(template)
        self.parse_detector_section(template)
        self.parse_sample_section(template)
        self.parse_other_sections(template)

        debugging = False
        if debugging is True:
            for keyword, value in template.items():
                print(f"{keyword}, {value}")
        return template



"""Writes a NeXus HDF5 file using h5py"""

import datetime
import h5py  # HDF5 support


RAW_MR_SCAN = """
17.92608    1037
17.92591    1318
17.92575    1704
17.92558    2857
17.92541    4516
17.92525    9998
17.92508    23819
17.92491    31662
17.92475    40458
17.92458    49087
17.92441    56514
17.92425    63499
17.92408    66802
17.92391    66863
17.92375    66599
17.92358    66206
17.92341    65747
17.92325    65250
17.92308    64129
17.92291    63044
17.92275    60796
17.92258    56795
17.92241    51550
17.92225    43710
17.92208    29315
17.92191    19782
17.92175    12992
17.92158    6622
17.92141    4198
17.92125    2248
17.92108    1321
"""


if __name__ == "__main__":
    print("Write a NeXus HDF5 file")
    fileName = "simple_example_test.nexus.hdf5"
    timestamp = datetime.datetime.now().astimezone().isoformat()

    # prepare the data
    data = {"mr": [], "I00": []}
    buffer = RAW_MR_SCAN.strip().split("\n")
    for row in buffer:
        (x, y) = row.split()
        data["mr"].append(float(x))
        data["I00"].append(float(y))

    # create the HDF5 NeXus file
    with h5py.File(fileName, "w") as f:
        f.attrs["file_name"] = fileName
        f.attrs["creator"] = "Pete R. Jemian <jemian@anl.gov> using h5py"
        f.attrs["HDF5_Version"] = h5py.version.hdf5_version
        f.attrs["NeXus_version"] = "4.2.1"
        f.attrs["h5py_version"] = h5py.version.version
        f.attrs["file_time"] = timestamp
        f.attrs["file_update_time"] = timestamp
        f.attrs["default"] = "entry"  # identify default NXentry group

        nxentry = f.create_group("entry")
        nxentry.attrs["NX_class"] = "NXentry"  # identify NeXus base class
        nxentry.attrs["default"] = "mr_scan"  # identify default NXdata group

        # store the scan data
        nxdata = nxentry.create_group("mr_scan")
        nxdata.attrs["NX_class"] = "NXdata"  # identify NeXus base class
        nxdata.attrs["signal"] = "I00"  # identify default data to plot
        nxdata.attrs["axes"] = "mr"  # identify default dimension scale to plot

        mr = nxdata.create_dataset("mr", data=data["mr"])
        mr.attrs["units"] = "degrees"

        i00 = nxdata.create_dataset("I00", data=data["I00"])
        i00.attrs["units"] = "counts"

        # fill in some optional metadata
        nxentry.create_dataset("title", data="APS USAXS instrument MR (alignment) scan")
        nxentry.create_dataset("start_time", data="2010-04-25T10:20:56-0500")
        nxentry.create_dataset("end_time", data="2010-04-25T10:21:16-0500")
        nxentry.create_dataset(
            "experiment_identifier", data="spec file 04_25.dat, scan #8"
        )
        nxentry.create_dataset(
            "experiment_description",
            data="alignment scan of the USAXS collimating optics",
        )
